from nose2 import events
from nose2.plugins.loader import discovery


class RemoveDuplicatedTests(events.Plugin, discovery.Discoverer):
    """If the DiscoveryLoader and the EggDiscoveryLoader plugins are enabled at the same time,
    there will be duplicated tests discovered as they both call _find_tests_in_module() which
    will discover tests no matter it is egg or not.
    """

    alwaysOn = True
    configSection = "discovery"

    def registerInSubprocess(self, event):
        event.pluginClasses.append(self.__class__)

    def _removeDuplicate(self, tests):
        uniqueTests = []
        for t in tests:
            if t not in uniqueTests:
                uniqueTests.append(t)
        return uniqueTests

    def loadTestsFromName(self, event):
        # This is where the EggDiscoveryLoader plugin introduce the duplicated plugin.
        # To-Do: Lodge the issue in nose2 repro.
        event.extraTests = self._removeDuplicate(event.extraTests)

    def loadTestsFromNames(self, event):
        return None



import sys
import logging
from iutest import dependencies
from iutest.core import pathutils

logger = logging.getLogger(__name__)

def isReimportFeatureAvailable(silentCheck=False):
    if not hasattr(sys, "maxint"):
        return False
    if silentCheck:
        return dependencies.ReimportWrapper.get().isValid()

    return dependencies.ReimportWrapper.check()


def isModuleModified(dotPath):
    if not isReimportFeatureAvailable():
        return None
    return dependencies.ReimportWrapper.getModule().modified(dotPath)


def reimportByModulePath(dotPath):
    if not isReimportFeatureAvailable():
        return
    try:
        mod = pathutils.objectFromDotPath(dotPath)
        dependencies.ReimportWrapper.getModule().reimport(mod)
        logger.info("Reimported module: %s", dotPath)
    except:
        logger.exception("Error reloading module: %s", dotPath)


def reimportAllChangedPythonModules(inclusiveKeyword=None, exclusiveKeyword=None):
    """
    Args:
        filterKeyword (str): A keyword to 
    """
    if not isReimportFeatureAvailable():
        return

    def _iterFilteredModules(modules):
        if not modules:
            return
            
        for module in modules:
            if inclusiveKeyword and inclusiveKeyword.lower() not in module.lower():
                logger.debug("Ignore %s since it does not match %s", module, inclusiveKeyword)
                continue

            if exclusiveKeyword and exclusiveKeyword.lower() in module.lower():
                logger.debug("Ignore %s since it matches %s", module, exclusiveKeyword)
                continue

            yield module

    changed = dependencies.ReimportWrapper.getModule().modified()
    changed = list(_iterFilteredModules(changed))
    if changed:
        print("Reimporting: {} ...".format(changed))
        successCount = 0
        failedCount = 0
        for module in changed:
            try:
                dependencies.ReimportWrapper.getModule().reimport(module)
            except Exception:
                logger.exception("Unable to reimport module %s", module)
                failedCount += 1
            else:
                successCount += 1
        if not failedCount:
            print("{} modules reimported.".format(successCount))
        else:
            if not successCount:
                print("All {} modules failed to reimport.".format(failedCount))
            else:
                print(
                    "{} modules reimported, {} modules failed to reimport.".format(
                        successCount, failedCount
                    )
                )

    else:
        print("No modules need to reimport :)")
    return changed



import collections
import logging
from iutest.core import constants
from iutest.qt import QtCore, variantToPyValue

logger = logging.getLogger(__name__)


def get():
    return AppSettings.get()


class SettingsGroupContext(object):
    def __init__(self, groupName):
        self._groupName = groupName
        self._settings = AppSettings.get()

    def __enter__(self):
        self._settings.beginGroup(self._groupName)
        return self._settings

    def __exit__(self, *_, **__):
        self._settings.endGroup()


class AppSettings(object):
    _instance = None

    @classmethod
    def get(cls):
        if not cls._instance:
            cls._instance = AppSettings()
        return cls._instance

    @staticmethod
    def _createIniSettings():
        return QtCore.QSettings(
            QtCore.QSettings.IniFormat,
            QtCore.QSettings.UserScope,
            constants.ORG_MGLAND,
            constants.APP_NAME,
        )

    def __init__(self):
        self._settings = self._createIniSettings()

    def saveSimpleConfig(self, key, value, sync=True):
        if isinstance(value, bool):
            value = 1 if value else 0

        self._qsettings().setValue(key, value)
        if sync:
            self._qsettings().sync()

    def simpleConfigValue(self, key, defaultValue=None):
        value = self._qsettings().value(key, defaultValue)
        return variantToPyValue(value)

    def simpleConfigStrValue(self, key, defaultValue=""):
        return str(self.simpleConfigValue(key, defaultValue=defaultValue))

    def simpleConfigBoolValue(self, key, defaultValue=False):
        value = self.simpleConfigValue(key, defaultValue=defaultValue)
        if value == "0":
            value = False
        return bool(value)

    def simpleConfigIntValue(self, key, defaultValue=0):
        configValue = self.simpleConfigValue(key, defaultValue=defaultValue) or 0
        if isinstance(configValue, str):
            if not configValue.isdigit():
                return 0
                
        return int(configValue)

    def removeConfig(self, key):
        self._qsettings().remove(key)

    def _qsettings(self):
        return object.__getattribute__(self, "_settings")

    def __getattribute__(self, attr):
        """We deligate everything else to QSettings object.
        """
        if hasattr(AppSettings, attr):
            return object.__getattribute__(self, attr)

        settingsObj = object.__getattribute__(self, "_qsettings")()
        return getattr(settingsObj, attr)

    def testDirSettings(self):
        with SettingsGroupContext(constants.CONFIG_KEY_SAVED_TEST_DIR) as settings:
            logger.debug("Settings ini file:%s", settings.fileName())
            names = settings.childGroups()
            data = collections.OrderedDict()
            for n in names:
                with SettingsGroupContext(n):
                    data[n] = (
                        settings.simpleConfigStrValue(
                            constants.CONFIG_KEY_TEST_TOP_DER
                        ),
                        settings.simpleConfigStrValue(
                            constants.CONFIG_KEY_TEST_START_DER
                        ),
                    )

        return data

    def __repr__(self):
        return "AppSettings({})".format(self._qsettings().fileName())



from iutest.core.runners import runnerconstants

RUNNER_REGISTRY = {
    runnerconstants.RUNNER_PYUNIT: "iutest.core.runners.pyunitrunner.PyUnitRunner",
    runnerconstants.RUNNER_NOSE2: "iutest.core.runners.nose2runner.Nose2TestRunner",
    runnerconstants.RUNNER_PYTEST: "iutest.core.runners.pytestrunner.PyTestTestRunner",
}


def getRunnerModes():
    """Get all runners in a sorted order.
    """
    return sorted(RUNNER_REGISTRY.keys())



import unittest
import os
import sys
import tempfile
import shutil

from iutest.core import importutils

@unittest.skipUnless(importutils.isReimportFeatureAvailable(silentCheck=True), "reimport feature is unavailable.")
class ImportUtilsTestCase(unittest.TestCase):
    def setUp(self):
        self._tempDir = tempfile.mkdtemp()
        self._varName = "myVariable"
        self._modName = "_testReimportModule"
        if self._tempDir not in sys.path:
            sys.path.append(self._tempDir)

    def _testModFilePath(self):
        return os.path.join(self._tempDir, (self._modName+".py"))

    def _writeModule(self, value):
        value = "'{}'".format(value) if isinstance(value, str) else value
        content = "{}={}".format(self._varName, value)
        with open(self._testModFilePath(), "w") as f:
            f.write(content)

    def tearDown(self):
        mod = __import__(self._modName)
        del mod
        if self._modName in sys.modules:
            del sys.modules[self._modName]

        if os.path.isdir(self._tempDir):
            shutil.rmtree(self._tempDir)

        if self._tempDir in sys.path:
            sys.path.remove(self._tempDir)

    def _getVarValue(self):
        mod = __import__(self._modName)
        return getattr(mod, self._varName)
    
    @unittest.skip("Looks like different modifications happen too closed for reimport to know it is modified.")
    def test_reimportByModulePath(self):
        action = lambda:importutils.reimportByModulePath(self._modName)
        self._checkReimport(action)

    def _checkReimport(self, reimportAction):
        oldValue = 1234
        self._writeModule(oldValue)
        self.assertEqual(self._getVarValue(), oldValue)

        for v in (0, "abc", True):
            self._writeModule(v)
            self.assertTrue(importutils.isModuleModified(self._modName))
            self.assertEqual(self._getVarValue(), oldValue)
            reimportAction()
            self.assertEqual(self._getVarValue(), v)
            oldValue = v

    @unittest.skip("Looks like different modifications happen too closed for reimport to know it is modified.")
    def test_reimportAllChangedPythonModules(self):
        action = lambda:importutils.reimportAllChangedPythonModules(inclusiveKeyword=self._modName)
        self._checkReimport(action)



import logging
import sys
from unittest import runner

from iutest.core import constants
from iutest.core import uistream
from iutest.plugins.pyunitextentions import pyunitcommon

logger = logging.getLogger(__name__)

class PyUnitTestRunnerWrapper(runner.TextTestRunner):
    def __init__(
        self, 
        verbosity=2, 
        failfast=False, 
        partialMode=constants.RUN_TEST_FULL
        ):
        self._verbosity = verbosity
        self.stream = uistream.UiStream()
        self._partialMode = partialMode

        # call baseClass.__init__() here means running the test:
        runner.TextTestRunner.__init__(
            self, 
            stream=self.stream,
            resultclass=PyUnitTestResult,
            verbosity=verbosity, 
            failfast=failfast)
    
    @staticmethod
    def _dummyFunction(*_, **__):
        pass
    
    def resetUiStreamResult(self):
        self.stream.setResult()

    def _getTestCaseFromTest(self, test):
        while not hasattr(test, "id"):
            tests = list(test)
            if not tests:
                logger.error("No test to run.")
                return
            test = tests[0]
            
        return test

    def run(self, test):
        # If we run test partially, we guarante that suite only contains a single testCase object.
        if self._partialMode != constants.RUN_TEST_FULL:
            test = self._getTestCaseFromTest(test)
            test.tearDown = self._dummyFunction
            if self._partialMode == constants.RUN_TEST_SETUP_ONLY:
                methodName = test.id().split(".")[-1]
                setattr(test, methodName, self._dummyFunction)
                logger.info("Run %s.setUp() only:", test.__class__.__name__)

            elif self._partialMode == constants.RUN_TEST_NO_TEAR_DOWN:
                logger.info("Skipped %s.tearDown():", test.__class__.__name__)

        return runner.TextTestRunner.run(self, test)


class PyUnitTestResult(runner.TextTestResult, pyunitcommon.PyUnitUiMixin):
    """A test result class that can show text results in ui, both in tree view and the log browser.
    """
    _originalStdOut = sys.stdout
    _originalStdErr = sys.stderr
    def __init__(self, stream, descriptions, verbosity):
        self.Cls = self.__class__
        self.Base = super(PyUnitTestResult, self)
        self.Base.__init__(stream, descriptions, verbosity)
        pyunitcommon.PyUnitUiMixin.__init__(self, stream)

    def startTestRun(self):
        self._atStartTestRun()
        self.Base.startTestRun()

    def startTest(self, test):
        self._atStartTest(test)
        with self.stream.linkInfoCtx(self._linkInfoFromTest(test)):            
            self.Base.startTest(test)

        self._startLogProcessers()
    
    def stopTest(self, test):
        self.Base.stopTest(test)
        self._atStopTest(test)

    def stopTestRun(self):
        self.Base.stopTestRun()
        self._atStopTestRun()
        resultCode = constants.TEST_RESULT_PASS if self.wasSuccessful() \
            else constants.TEST_RESULT_ERROR
        self.stream.setResult(resultCode)

    def addSuccess(self, test):
        with self.stream.resultCtx(constants.TEST_RESULT_PASS):
            self.Base.addSuccess(test)
            self._atOutcomeAvailable(test.id(), constants.TEST_RESULT_PASS)

    def addError(self, test, err):
        with self.stream.resultCtx(constants.TEST_RESULT_ERROR):
            self.Base.addError(test, err)
            self._atOutcomeAvailable(test.id(), constants.TEST_RESULT_ERROR)

    def addFailure(self, test, err):
        with self.stream.resultCtx(constants.TEST_RESULT_FAIL):
            self.Base.addFailure(test, err)
            self._atOutcomeAvailable(test.id(), constants.TEST_RESULT_FAIL)

    def addSkip(self, test, reason):
        with self.stream.resultCtx(constants.TEST_RESULT_SKIP):
            self.Base.addSkip(test, reason)
            self._atOutcomeAvailable(test.id(), constants.TEST_RESULT_SKIP)

    def addExpectedFailure(self, test, err):
        with self.stream.resultCtx(constants.TEST_RESULT_EXPECTED_FAIL):
            self.Base.addExpectedFailure(test, err)
            self._atOutcomeAvailable(test.id(), constants.TEST_RESULT_EXPECTED_FAIL)

    def addUnexpectedSuccess(self, test):
        with self.stream.resultCtx(constants.TEST_RESULT_UNEXPECTED_PASS):
            self.Base.addUnexpectedSuccess(test)
            self._atOutcomeAvailable(test.id(), constants.TEST_RESULT_UNEXPECTED_PASS)

    def printErrorList(self, flavour, errors):
        for test, err in errors:
            with self.stream.resultCtx(constants.TEST_RESULT_FAIL): 
                self.stream.writeln(self.separator1)

            with self.stream.linkInfoCtx(self._linkInfoFromTest(test)):
                self.stream.writeln("%s: %s" % (flavour,self.getDescription(test)))

            self.stream.writeln(self.separator2)

            with self.stream.processStackTraceLinkCtx():
                self.stream.writeln("%s" % err)
        
        # We set the result code to error, so the summary will be in red.
        if errors:
            self.stream.setResult(constants.TEST_RESULT_ERROR)



import unittest
import logging

logger = logging.Logger(__name__)

def _nose2Params(*paramList):
    """Copy this directly from nose2 so that the tests here can still run without nose2.
    """
    def decorator(func):
        func.paramList = paramList
        return func
    return decorator


class DummyTests(unittest.TestCase):
    """Some dummy tests for the IUTest UI manual tests.
    """

    def setUp(self):
        print("Call {}.setUp()".format(self.__class__.__name__))

    def tearDown(self):
        print("Call {}.tearDown()".format(self.__class__.__name__))

    def test_passed(self):
        logger.info("About to pass.")
        self.assertTrue(True)

    @unittest.skip("Test the skipped tests.")
    def test_skipped(self):
        pass

    def test_error(self):
        logger.warning("About to have test error.")
        raise RuntimeError("A test error.")

    def test_failed(self):
        self.assertTrue(False, "False will never be True.")

    @unittest.expectedFailure
    def test_unexpectedPass(self):
        logger.warning("This should not pass.")
        self.assertTrue(True)

    @unittest.expectedFailure
    def test_expectedFailure(self):
        logger.warning("This should not pass.")
        self.assertTrue(False)

    @unittest.skipUnless(
        hasattr(unittest.TestCase, "subTest"),
        "Current TestCase version does not support subTest",
    )
    def test_subTests(self):
        for i in range(0, 6):
            with self.subTest(i=i):
                self.assertEqual(i % 2, 0)

    @_nose2Params((1, 2), (2, 3), (6, 5), (4, 6))
    def test_parameters(self, num1, num2):
        logger.info("Test with %s < %s", num1, num2)
        self.assertLess(num1, num2)



import unittest

from iutest.qt import QtWidgets
from iutest import dcc


class AppModeTestCase(unittest.TestCase):
    @unittest.skipUnless(dcc.isInsideMaya(), "It is maya only test")
    def test_findParentMayaWindow(self):
        window = dcc.findParentWindow()
        self.assertTrue(window)
        self.assertTrue(isinstance(window, QtWidgets.QWidget))
        self.assertTrue("maya" in window.objectName().lower())

    @unittest.skipUnless(dcc.isStandalone(), "It is standalone only test")
    def test_findParentWindow(self):
        window = dcc.findParentWindow()
        self.assertFalse(window)



class TestRunInfo(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self.runTestIds = []
        self.failedTestId = None
        self.runCount = 0
        self.successCount = 0
        self.failedCount = 0
        self.errorCount = 0
        self.skipCount = 0
        self.expectedFailureCount = 0
        self.unexpectedSuccessCount = 0

        self._sessionStartTime = 0
        self.sessionRunTime = 0

        self._testStartTimes = {}
        self.singleTestRunTime = 0


import collections

ORG_MGLAND = "MGLAND"
APP_NAME = "IUTest"

TEST_RESULT_NONE = 0
TEST_RESULT_RUNNING = 1
TEST_RESULT_SKIP = 2
TEST_RESULT_PASS = 3
TEST_RESULT_EXPECTED_FAIL = 4
TEST_RESULT_UNEXPECTED_PASS = 5
TEST_RESULT_FAIL = 6
TEST_RESULT_ERROR = 7

RUN_TEST_SETUP_ONLY = 0
RUN_TEST_NO_TEAR_DOWN = 1
RUN_TEST_FULL = 2

ITEM_CATEGORY_ALL = 0
ITEM_CATEGORY_PACKAGE = 1
ITEM_CATEGORY_MODULE = 2
ITEM_CATEGORY_SUITE = 3
ITEM_CATEGORY_TEST = 4


KEYWORD_TEST_STATE_NONE = ":normal"
KEYWORD_TEST_STATE_RUN = ":ran"
KEYWORD_TEST_STATE_SKIPPED = ":skipped"
KEYWORD_TEST_STATE_SUCCESS = ":success"
KEYWORD_TEST_STATE_EXPECTED_FAIL = ":error"
KEYWORD_TEST_STATE_UNEXPECTED_PASS= ":error"
KEYWORD_TEST_STATE_FAILED = ":failed"
KEYWORD_TEST_STATE_ERROR = ":error"
KEYWORD_TEST_STATES = collections.OrderedDict(
    [
        (KEYWORD_TEST_STATE_NONE, TEST_RESULT_NONE),
        (KEYWORD_TEST_STATE_RUN, -1),
        (KEYWORD_TEST_STATE_SKIPPED, TEST_RESULT_SKIP),
        (KEYWORD_TEST_STATE_SUCCESS, TEST_RESULT_PASS),
        (KEYWORD_TEST_STATE_EXPECTED_FAIL, TEST_RESULT_EXPECTED_FAIL),
        (KEYWORD_TEST_STATE_UNEXPECTED_PASS, TEST_RESULT_UNEXPECTED_PASS),
        (KEYWORD_TEST_STATE_FAILED, TEST_RESULT_EXPECTED_FAIL),
        (KEYWORD_TEST_STATE_ERROR, TEST_RESULT_ERROR),
    ]
)

TEST_ICON_NONE_SUFFIX = ""
TEST_ICON_RUNNING_SUFFIX = "_running"
TEST_ICON_SKIP_SUFFIX = "_skipped"
TEST_ICON_SUCCESS_SUFFIX = "_success"
TEST_ICON_EXPECTED_FAIL_SUFFIX = "_expectedFailure"
TEST_ICON_UNEXPECTED_PASS_SUFFIX = "_unepectedSuccess"
TEST_ICON_FAILED_SUFFIX = "_failed"
TEST_ICON_ERROR_SUFFIX = "_error"
TEST_ICON_SUFFIXES = (
    TEST_ICON_NONE_SUFFIX,
    TEST_ICON_RUNNING_SUFFIX,
    TEST_ICON_SKIP_SUFFIX,
    TEST_ICON_SUCCESS_SUFFIX,
    TEST_ICON_EXPECTED_FAIL_SUFFIX,
    TEST_ICON_UNEXPECTED_PASS_SUFFIX,
    TEST_ICON_FAILED_SUFFIX,
    TEST_ICON_ERROR_SUFFIX,
)


PANEL_VIS_STATE_RIGHT_ON = 0
PANEL_VIS_STATE_LEFT_ON = 1
PANEL_VIS_STATE_BOTH_ON = 2
PANEL_VIS_STATE_ICON_SUFFIXES = ("_right", "_left", "_both")


# Config Keys --------------------------------------------
CONFIG_KEY_SAVED_TEST_DIR = "testDirs"
CONFIG_KEY_TEST_TOP_DER = "topDir"
CONFIG_KEY_TEST_START_DER = "testRootDir"

CONFIG_KEY_LAST_TEST_ROOT_DIR = "lastTestRootDir"
CONFIG_KEY_LAST_TOP_DIR = "lastTopDir"

CONFIG_KEY_PANEL_VIS_STATE = "panelVisState"
CONFIG_KEY_AUTO_FILTERING_STATE = "autoFiltering"
CONFIG_KEY_AUTO_CLEAR_LOG_STATE = "autoClearLog"

CONFIG_KEY_STOP_ON_ERROR = "stopOnError"

CONFIG_KEY_CODE_EDITOR = "codeEditor"
CODE_FILE_VAR = "$file"
CODE_LINE_VAR = "$line"
CONFIG_KEY_CODE_EDITOR_DEFAULT = 'code --goto "{}":{}'.format(
    CODE_FILE_VAR, CODE_LINE_VAR
)

CONFIG_KEY_LAST_RUNNER_MODE = "lastRunnerMode"



from iutest.core import gotocode
from iutest.ui import uiconstants
from iutest.qt import QtGui, QtWidgets


class LogBrowser(QtWidgets.QTextBrowser):
    def __init__(self, parent=None):
        QtWidgets.QTextBrowser.__init__(self, parent)
        fn = self.font()
        pixelSize = max(12, fn.pixelSize() + 2)
        fn.setPixelSize(pixelSize)
        fn.setFamily("Courier New")
        self.setFont(fn)
        self.setReadOnly(True)
        self.setWordWrapMode(QtGui.QTextOption.NoWrap)
        self.setOpenLinks(False)
        self.anchorClicked.connect(self.onLinkClicked)
        self._initCSS()
        self._codeVisitor = gotocode.CodeLineVisitor(self)

    def _initCSS(self):
        css = """
        a:link {
            color: #C8C8C8;
        }

        * {
            white-space: pre;
        }
        """
        # text-decoration: none;
        self.document().setDefaultStyleSheet(css)

    def logWithColor(self, msg, color, *args):
        self.moveCursor(QtGui.QTextCursor.End)
        # SELF.setTextColor(color)
        msg = msg if not args else (msg % args)
        msg = "<font color=%s>%s</font>" % (color.name(), msg)
        self.insertHtml(msg)
        self.moveCursor(QtGui.QTextCursor.End)
        self.horizontalScrollBar().setValue(0)

    def onLinkClicked(self, url):
        path, _, line = url.query().rpartition("=")
        if not path:
            return
        if not line:
            line = 0
        self._codeVisitor.goTo(path, line)

    def logSeparator(self,):
        sep = ">" * 70
        self.logWarning("<br>{}<br>".format(sep))

    def logInformation(self, msg, *args):
        self.logWithColor(msg, uiconstants.LOG_COLOR_INFORMATION, *args)

    def logSuccess(self, msg, *args):
        self.logWithColor(msg, uiconstants.LOG_COLOR_SUCCESS, *args)

    def logFailed(self, msg, *args):
        self.logWithColor(msg, uiconstants.LOG_COLOR_FAILED, *args)

    def logError(self, msg, *args):
        self.logWithColor(msg, uiconstants.LOG_COLOR_ERROR, *args)

    def logWarning(self, msg, *args):
        self.logWithColor(msg, uiconstants.LOG_COLOR_WARNING, *args)



import logging

logger = logging.getLogger(__name__)


class _ErrorDummy(object):
    def __getattribute__(self, name):
        return _ErrorDummy()

    def __call__(self, *_, **__):
        return _ErrorDummy()

    def __repr__(self):
        return "Error Happened."

    def __iter__(self):
        yield _ErrorDummy()

    def __getitem__(self, index):
        return _ErrorDummy()

    def __bool__(self):
        return False

    def __nonzero__(self):
        return False


class _DependencyWrapper(object):
    @classmethod
    def get(cls, silent=False):
        """Get an instance of the wrapper object.

        Args:
            silent (bool): Whether we issue errors or debug when the dependency 
                is not installed.
        """
        if not hasattr(cls, "_instance"):
            return None

        if not cls._instance:
            cls._instance = cls()
        return cls._instance

    @classmethod
    def getModule(cls, silent=False):
        wrapper = cls.get(silent=silent)
        return wrapper._mod if wrapper._mod else _ErrorDummy()

    @classmethod
    def reload(cls, silent=True):
        """Try reimport the dependency module.

        Args:
            silent (bool): Whether we issue errors or debug when the dependency 
                is not installed.
        """
        cls.get()._tryImport(force=True, silent=silent)

    def __init__(self):
        self._mod = None
        self._tryImport(force=False, silent=True)

    @classmethod
    def _issueNotInstalledError(cls, silent=True):
        if not silent:
            logger.error("The package '%s' is not installed", cls.name())
        else:
            logger.debug("The package '%s' is not installed", cls.name())

    @classmethod
    def _issueNotImplementedError(cls):
        err = "Please use a derived class instead of base class {}".format(cls.__name__)
        raise NotImplementedError(err)

    def _tryImport(self, force, silent):
        self._issueNotImplementedError()

    @classmethod
    def name(cls):
        cls._issueNotImplementedError()

    def isValid(self):
        return bool(self._mod)

    @classmethod
    def check(cls):
        if not cls.get().isValid():
            cls._issueNotInstalledError(silent=False)
            return False
        return True


class ReimportWrapper(_DependencyWrapper):
    _instance = None

    @classmethod
    def name(cls):
        return "reimport"

    def _tryImport(self, force, silent):
        if not force and self._mod:
            return
        self._mod = None
        try:
            import reimport

            self._mod = reimport
        except ImportError:
            self._issueNotInstalledError(silent)


class Nose2Wrapper(_DependencyWrapper):
    _instance = None

    @classmethod
    def name(cls):
        return "nose2"

    def _tryImport(self, force, silent):
        if not force and self._mod:
            return
        self._mod = None
        try:
            import nose2

            self._mod = nose2
        except ImportError:
            self._issueNotInstalledError(silent)


class PyTestWrapper(_DependencyWrapper):
    _instance = None

    @classmethod
    def name(cls):
        return "pytest"

    def _tryImport(self, force, silent):
        if not force and self._mod:
            return
        self._mod = None
        try:
            import pytest

            self._mod = pytest
        except ImportError:
            self._issueNotInstalledError(silent)



import logging

from iutest.qt import QtCore, QtWidgets, Signal, variantToPyValue, iconFromPath
from iutest.core import iconutils
from iutest.core import constants
from iutest.core import pathutils
from iutest.core import gotocode
from iutest.core import importutils
from iutest.core import loggingutils
from iutest.core import uistream

logger = logging.getLogger(__name__)


class ViewStates(object):
    def __init__(self, testTreeView):
        self._treeWidget = testTreeView
        self.reset()

    def reset(self):
        self._currentItemId = None
        self._itemExpandStates = {}
        self._horizontalScrollValue = 0
        self._verticalScrollValue = 0

    def save(self):
        if not self._treeWidget:
            return

        self.reset()
        currentItem = self._treeWidget.currentItem()
        if currentItem:
            self._currentItemId = self._treeWidget.testIdOfItem(currentItem)

        for testId, item in self._treeWidget.allItemsIdMap().items():
            self._itemExpandStates[testId] = item.isExpanded()
        self._horizontalScrollValue = self._treeWidget.horizontalScrollBar().value()
        self._verticalScrollValue = self._treeWidget.verticalScrollBar().value()

    def _restoreScrollValue(self):
        self._treeWidget.horizontalScrollBar().setValue(self._horizontalScrollValue)
        self._treeWidget.verticalScrollBar().setValue(self._verticalScrollValue)

    def restore(self):
        if not self._treeWidget:
            return
        itemIdMap = self._treeWidget.allItemsIdMap()
        cItem = itemIdMap.get(self._currentItemId, None)
        if cItem:
            self._treeWidget.setCurrentItem(cItem)
        for testId, expand in self._itemExpandStates.items():
            item = itemIdMap.get(testId, None)
            if item:
                item.setExpanded(expand)
        QtCore.QTimer.singleShot(0, self._restoreScrollValue)


class UnitTestTreeView(QtWidgets.QTreeWidget):
    runAllTest = Signal()
    runTests = Signal(tuple)
    runSetupOnly = Signal(str)
    runWithoutTearDown = Signal(str)

    _testAllIcons = []
    _testPackageIcons = []
    _testModuleIcons = []
    _testSuiteIcons = []
    _testCaseIcons = []
    _allItemIconSet = []

    _runPartialIcon = None
    _runAllIcon = None
    _runSelectedIcon = None
    _reimportAndRunIcon = None

    supportPartialCategories = (
        constants.ITEM_CATEGORY_SUITE,
        constants.ITEM_CATEGORY_TEST,
    )

    class _TreeItemDelegate(QtWidgets.QItemDelegate):
        def drawFocus(self, painter, styleOptionViewItem, rect):
            pass

    def __init__(self, parent):
        QtWidgets.QTreeWidget.__init__(self, parent)
        self._uiStream = uistream.UiStream()
        self._codeVisitor = gotocode.CodeLineVisitor(self)

        self.setItemDelegate(self._TreeItemDelegate(self))
        self.setTextElideMode(QtCore.Qt.ElideLeft)
        self.setColumnCount(2)
        self._setHeaderStretch()
        self.setHeaderHidden(True)
        self.setAlternatingRowColors(True)
        self.setExpandsOnDoubleClick(False)
        self.setSelectionMode(self.ExtendedSelection)
        self.itemDoubleClicked.connect(self.onItemDoubleClicked)
        self.setIndentation(10)

        self._rootTestItem = None
        self._testCases = []
        self._allItemsIdMap = {}
        self._initAllIcons()

        self._testManager = None
        self._viewStates = ViewStates(self)

        self._contextMenu = QtWidgets.QMenu(self)
        self._makeContextMenu()

        self.setToolTip(
            "This view lists out the tests, <b>double click</b> on them to run them, <b>Ctrl+C</b> to copy the python module path."
        )

    def setTestManager(self, manager):
        self._testManager = manager

    def _setHeaderStretch(self):
        header = self.header()
        header.setStretchLastSection(False)
        if hasattr(header, "setSectionResizeMode"):
            header.setSectionResizeMode(0, header.Stretch)
            header.setSectionResizeMode(1, header.ResizeToContents)
        else:
            header.setResizeMode(0, header.Stretch)
            header.setResizeMode(1, header.ResizeToContents)

        header.resizeSection(1, 10)

    def _iterAllDescendentItem(self, parentItem, category):
        for i in range(parentItem.childCount()):
            item = parentItem.child(i)
            if self._categoryOfItem(item) == category:
                yield item
                continue

            for it in self._iterAllChildrenItem(item, category):
                yield it

    def _splitPackageTestIds(self, *testIds):
        if not self._testManager.avoidRunTestsOnPackageLevel():
            return tuple(testIds)

        items = filter(None, [self._findItemById(tid) for tid in testIds])
        return self._splitPackageTestItems(*items)

    def _splitPackageTestItems(self, *items):
        if not self._testManager.avoidRunTestsOnPackageLevel():
            return tuple([self.testIdOfItem(item) for item in items])

        testIds = []
        for item in items:
            if self._categoryOfItem(item) != constants.ITEM_CATEGORY_PACKAGE:
                testIds.append(self.testIdOfItem(item))
                continue
            
            splittedTests = [self.testIdOfItem(item) for item in \
                self._iterAllDescendentItem(item, constants.ITEM_CATEGORY_MODULE)]
            testIds.extend(splittedTests)
        return tuple(testIds)

    def onItemDoubleClicked(self, item, *_, **__):
        if item is self._rootTestItem:
            self.runAllTest.emit()
            return

        self.runTests.emit(self._splitPackageTestItems(item))

    @classmethod
    def _initComboIcons(cls, iconVarName, iconFileName):
        iconPaths = iconutils.iconPathSet(iconFileName, constants.TEST_ICON_SUFFIXES)
        icons = []
        for path in iconPaths:
            icons.append(iconFromPath(path))
        setattr(cls, iconVarName, icons)

    @classmethod
    def _initAllIcons(cls):
        if cls._allItemIconSet:
            return

        cls._initComboIcons("_testAllIcons", "all.svg")
        cls._initComboIcons("_testPackageIcons", "package.svg")
        cls._initComboIcons("_testModuleIcons", "module.svg")
        cls._initComboIcons("_testSuiteIcons", "testSuite.svg")
        cls._initComboIcons("_testCaseIcons", "testCase.svg")
        cls._allItemIconSet = (
            cls._testAllIcons,
            cls._testPackageIcons,
            cls._testModuleIcons,
            cls._testSuiteIcons,
            cls._testCaseIcons,
        )
        iconutils.initSingleClassIcon(cls,"_reimportAndRunIcon", "reimportAndRerun.svg")
        iconutils.initSingleClassIcon(cls,"_runPartialIcon", "run_partial.svg")
        iconutils.initSingleClassIcon(cls,"_runAllIcon", "runAll.svg")
        iconutils.initSingleClassIcon(cls,"_runSelectedIcon", "runSelected.svg")

    def setFilterKeywords(self, keywords, ensureFirstMatchVisible=False):
        itemStates = {}
        if keywords:
            itemToFocus = None
            for item in self._testCases:
                self._accumulateItemVisibility(item, keywords, itemStates)
                if not itemToFocus:
                    itemToFocus = item

            # self._debugItemVisStates(itemStates)

            for index in range(self._rootTestItem.childCount()):
                self._recursivelySetItemVisibility(
                    self._rootTestItem.child(index), itemStates, state=None
                )

            if itemToFocus and ensureFirstMatchVisible:
                self.focusItem(itemToFocus)
        else:
            for index in range(self._rootTestItem.childCount()):
                self._recursivelySetItemVisibility(
                    self._rootTestItem.child(index), itemStates, state=True
                )

    def _recursivelySetItemVisibility(self, item, itemStates, state=None):
        visible = state
        stateKey = self.testIdOfItem(item)
        if state is None:
            visible = itemStates.get(stateKey, True)

        if not visible:
            item.setHidden(True)
            return

        item.setHidden(False)
        for index in range(item.childCount()):
            self._recursivelySetItemVisibility(item.child(index), itemStates, state)

    def _itemMatchs(self, item, keywords):
        state = variantToPyValue(item.data(0, QtCore.Qt.UserRole + 1))
        stateMatch = False
        txtMatch = True
        lbl = self.testIdOfItem(item).lower()
        hasFilter = False
        for each in keywords:
            if not each.startswith(":"):
                if each not in lbl:
                    return False

            if each.startswith(":"):
                hasFilter = True
                if not stateMatch:
                    desiredState = constants.TEST_RESULT_NONE
                    if each in constants.KEYWORD_TEST_STATES:
                        desiredState = constants.KEYWORD_TEST_STATES[each]
                    if desiredState < 0:
                        stateMatch = state > 0
                    else:
                        stateMatch = state == desiredState

        stateMatch = stateMatch or not hasFilter
        return stateMatch and txtMatch

    def _accumulateItemVisibility(self, item, keywords, itemStates, state=None):
        if not item:
            return

        if item is self._rootTestItem:
            return

        stateKey = self.testIdOfItem(item)
        itemStates.setdefault(stateKey, state)
        if state:
            itemStates[stateKey] = state
        else:
            state = self._itemMatchs(item, keywords)

            itemStates[stateKey] = itemStates[stateKey] or state
            state = itemStates[stateKey]

        self._accumulateItemVisibility(item.parent(), keywords, itemStates, state=state)

    def _debugItemVisStates(self, states):
        for tid in sorted(states.keys()):
            print(tid, states[tid])

    def _categoryOfItem(self, item):
        return variantToPyValue(item.data(0, QtCore.Qt.UserRole))

    def _stateOfItem(self, item):
        return variantToPyValue(item.data(0, QtCore.Qt.UserRole + 1))

    def _setItemIconState(self, item, state):
        self._initAllIcons()
        category = self._categoryOfItem(item)
        icons = self._allItemIconSet[category]
        item.setData(0, QtCore.Qt.UserRole + 1, state)
        item.setIcon(0, icons[state])

    def _setTestIconStateToDecendents(self, item, state):
        self._setItemIconState(item, state)
        for i in range(item.childCount()):
            self._setTestIconStateToDecendents(item.child(i), state)

    def _setTestIconStateToAncestors(self, item, state):
        self._setItemIconState(item, state)
        if item is not self._rootTestItem:
            self._setTestIconStateToAncestors(item.parent(), state)

    def resetAllItemsToNormal(self):
        self._setTestIconStateToDecendents(
            self._rootTestItem, constants.TEST_RESULT_NONE
        )
        self._resetExpandStates(self._rootTestItem)
        for item in self._testCases:
            self._resetItem(item, False)

    def _resetItem(self, item, applyToAllChildren=False):
        item.setText(1, "")
        if variantToPyValue(item.data(1, QtCore.Qt.UserRole)):
            item.setData(1, QtCore.Qt.UserRole, 0)

        if applyToAllChildren:
            for i in range(item.childCount()):
                self._resetItem(item.child(i), applyToAllChildren)

    def allItemsIdMap(self):
        return self._allItemsIdMap

    def _findItemById(self, testId):
        return self._allItemsIdMap.get(testId, None)

    def resetTestItemsById(self, testIds):
        for tid in testIds:
            item = self._findItemById(tid)
            if not item:
                continue

            self._resetItem(item, True)
            self._setTestIconStateToAncestors(item, constants.TEST_RESULT_NONE)
            self._setTestIconStateToDecendents(item, constants.TEST_RESULT_NONE)
        self.repaint()

    def focusItem(self, item):
        if item:
            self.clearSelection()
            self.setCurrentItem(item)
            self.scrollToItem(item)

    def _calculateAncestorItemStates(self, item, updatedIds):
        cId = self.testIdOfItem(item)
        if cId in updatedIds:
            return

        updatedIds.add(cId)
        stateToSet = constants.TEST_RESULT_NONE
        for i in range(item.childCount()):
            cState = variantToPyValue(item.child(i).data(0, QtCore.Qt.UserRole + 1))
            stateToSet = max(stateToSet, cState)
        
        self._setItemIconState(item, stateToSet)

        if item is self._rootTestItem:
            return

        self._calculateAncestorItemStates(item.parent(), updatedIds)

    def testIdOfItem(self, item):
        return str(item.toolTip(0))

    def _resetExpandStates(self, item):
        item.setExpanded(True)
        childCount = item.childCount()
        if childCount == 1:
            self._resetExpandStates(item.child(0))
        else:
            for i in range(childCount):
                item.child(i).setExpanded(False)

    def _firstSelectedModulePath(self):
        for item in self.selectedItems():
            return self.testIdOfItem(item)
        return None

    def copyFirstSelectedTestId(self):
        firstSelectedModulePath = self._firstSelectedModulePath()
        if firstSelectedModulePath:
            QtWidgets.QApplication.clipboard().setText(firstSelectedModulePath)
            logger.info("Module Path copied to clipboard: %s", firstSelectedModulePath)
        else:
            logger.info("No item selected to copy the module path.")

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_C and (
            event.modifiers() & QtCore.Qt.ControlModifier
        ):
            self.copyFirstSelectedTestId()
            event.accept()
            return

        QtWidgets.QTreeWidget.keyPressEvent(self, event)

    def reload(self, keepUiStates=True):
        if keepUiStates:
            self._viewStates.save()

        self._testCases = []
        self._rootTestItem = None
        self._allItemsIdMap = {}
        self.clear()

        startDirOrModule = self._testManager.startDirOrModule()
        if not startDirOrModule:
            return

        self._rootTestItem = QtWidgets.QTreeWidgetItem([startDirOrModule])
        self._rootTestItem.setToolTip(0, startDirOrModule)
        self._rootTestItem.setData(0, QtCore.Qt.UserRole, constants.ITEM_CATEGORY_ALL)
        self._setItemIconState(self._rootTestItem, constants.TEST_RESULT_NONE)
        self.addTopLevelItem(self._rootTestItem)

        self._allItemsIdMap[startDirOrModule] = self._rootTestItem

        isModule = not pathutils.isPath(startDirOrModule)
        headingCount = 0
        heading = ""
        if isModule:
            headingCount = len(startDirOrModule) + 1
            heading = startDirOrModule + "."
        
        testCount = 0
        for test in self._testManager.iterAllTestIds():
            testCount = testCount + 1
            if isModule and test.startswith(startDirOrModule):
                test = test[headingCount:]
            testPaths = test.split(".")
            cParent = self._rootTestItem
            for i, p in enumerate(testPaths):
                path = heading + ".".join(testPaths[0 : (i + 1)])
                if path in self._allItemsIdMap:
                    cParent = self._allItemsIdMap[path]
                    continue

                item = QtWidgets.QTreeWidgetItem(cParent)
                item.setText(0, p)
                item.setToolTip(0, path)
                item.setSizeHint(0, QtCore.QSize(20, 20))
                self._allItemsIdMap[path] = item
                cParent = item

            testCase = cParent
            testCase.setData(0, QtCore.Qt.UserRole, constants.ITEM_CATEGORY_TEST)
            self._setItemIconState(testCase, constants.TEST_RESULT_NONE)
            self._testCases.append(cParent)

            caseItem = testCase.parent()
            caseItem.setData(0, QtCore.Qt.UserRole, constants.ITEM_CATEGORY_SUITE)
            self._setItemIconState(caseItem, constants.TEST_RESULT_NONE)

            moduleItem = caseItem.parent()
            moduleItem.setData(0, QtCore.Qt.UserRole, constants.ITEM_CATEGORY_MODULE)
            self._setItemIconState(moduleItem, constants.TEST_RESULT_NONE)

            packageItem = moduleItem.parent()
            while packageItem is not self._rootTestItem and packageItem:
                packageItem.setData(
                    0, QtCore.Qt.UserRole, constants.ITEM_CATEGORY_PACKAGE
                )
                self._setItemIconState(packageItem, constants.TEST_RESULT_NONE)
                packageItem = packageItem.parent()
        
        self._uiStream.write("\n{}\n".format("-"*20))
        if not self._testManager.hasLastListerError():
            self._uiStream.write("{} tests collected.\n\n".format(testCount))
        else:
            with self._uiStream.resultCtx(constants.TEST_RESULT_ERROR):
                msg = "Error collecting tests from {}.\n\n".format(self._testManager.startDirOrModule())
                self._uiStream.write(msg)

        if keepUiStates:
            self._viewStates.restore()
        else:
            self._resetExpandStates(self._rootTestItem)

    def onSingleTestStart(self, testId, startTime):
        isParameterized, testId = self._testManager.parseParameterizedTestId(testId)
        item = self._findItemById(testId)
        if item:
            self.focusItem(item)
            self._setTestIconStateToAncestors(item, constants.TEST_RESULT_RUNNING)
            if not isParameterized or not variantToPyValue(
                item.data(1, QtCore.Qt.UserRole)
            ):
                item.setData(1, QtCore.Qt.UserRole, startTime)
                item.setText(1, "running...")

    def onSingleTestStop(self, testId, endTime):
        _, testId = self._testManager.parseParameterizedTestId(testId)
        item = self._findItemById(testId)
        if item:
            startTime = variantToPyValue(item.data(1, QtCore.Qt.UserRole))
            item.setData(1, QtCore.Qt.UserRole + 1, endTime)
            rep = "%.3f s" % (endTime - startTime)
            item.setText(1, rep)

    def showResultOnItemByTestId(self, testId, state):
        _, testId = self._testManager.parseParameterizedTestId(testId)
        item = self._findItemById(testId)
        if item:
            # For parameterized test, some might succeed but others might failed, we make sure
            # we set failed for this situation.
            if state == constants.TEST_RESULT_SKIP or state > variantToPyValue(
                self._stateOfItem(item)
            ):
                self._setItemIconState(item, state)

    def onAllTestsFinished(self):
        lastRunIds = self._testManager.lastRunTestIds()
        updatedIds = set()

        lastFailedItem = None

        for testId in lastRunIds:
            item = self._findItemById(testId)
            if not item:
                continue
            
            self._calculateAncestorItemStates(item.parent(), updatedIds)
            if self._testManager.lastFailedTestId() == testId:
                lastFailedItem = item

        self.focusItem(lastFailedItem)

    def testCount(self):
        return len(self._testCases)

    def hasTests(self):
        return bool(self._testCases)

    def selectedTestIds(self, decomposePackageIfNecessary=False):
        itemsByModulePath = {
            self.testIdOfItem(item): item for item in self.selectedItems()
        }

        lastKey = None
        testsToRun = []
        for key in sorted(itemsByModulePath.keys()):
            if lastKey and (key).startswith(lastKey):
                continue

            lastKey = key + "."
            logger.debug("Running test: %s", key)
            testsToRun.append(key)

        if decomposePackageIfNecessary:
            return self._splitPackageTestIds(*testsToRun)

        return tuple(testsToRun)

    def _iterSelectedSuiteOrTestCaseItems(self):
        for item in self.iterSelectedItemsOfCategories(*self.supportPartialCategories):
            yield item

    def _firstSelectedTestCaseItem(self):
        """Return first selected test case id.
        
        Notes:
            If there is any selected test cases, return the first id. 
            If there isn't but there is selected test suite, return its first test case id.
        """
        for item in self._iterSelectedSuiteOrTestCaseItems():
            if self._categoryOfItem(item) == constants.ITEM_CATEGORY_TEST:
                return item
            
            for i in range(item.childCount()):
                return item.child(i)

        return None

    def firstSelectedTestCaseId(self):
        item = self._firstSelectedTestCaseItem()
        return None if not item else self.testIdOfItem(item)

    def hasSelectedTests(self, hasSelectedTestOrCase=False):
        if not hasSelectedTestOrCase:
            return bool(self.selectionModel().hasSelection())

        return self.hasSelectionOfCategories(*self.supportPartialCategories)

    def hasSelectionOfCategories(self, *categories):
        for item in self.selectedItems():
            if self._categoryOfItem(item) in categories:
                return True
        return False

    def iterSelectedItemsOfCategories(self, *categories):
        for item in self.selectedItems():
            if self._categoryOfItem(item) in categories:
                yield item

    def _makeContextMenu(self):
        self._initAllIcons()

        self._runSetupOnlyAct = QtWidgets.QAction("Run setUp( ) Only", self)
        self._runSetupOnlyAct.triggered.connect(self._atRunSetupOnly)
        self._runSetupOnlyAct.setIcon(self._runPartialIcon)

        self._runWithoutTearDownAct = QtWidgets.QAction("Run Without tearDown( )", self)
        self._runWithoutTearDownAct.triggered.connect(self._atRunWithoutTearDown)
        self._runWithoutTearDownAct.setIcon(self._runPartialIcon)

        self._runSelectedAct = QtWidgets.QAction("Run Selected", self)
        self._runSelectedAct.triggered.connect(self._atRunSelected)
        self._runSelectedAct.setIcon(self._runSelectedIcon)

        self._runAllAct = QtWidgets.QAction("Run All", self)
        self._runAllAct.triggered.connect(self._atRunAll)
        self._runAllAct.setIcon(self._runAllIcon)

        self._copyPathAct = QtWidgets.QAction("Copy Selected Path", self)
        self._copyPathAct.triggered.connect(self.copyFirstSelectedTestId)

        self._goToCodeAct = QtWidgets.QAction("Go To Code...", self)
        self._goToCodeAct.triggered.connect(self._atGoToCode)

        self._reloadModulesAct = QtWidgets.QAction("Reload Selected Modules", self)
        self._reloadModulesAct.setVisible(importutils.isReimportFeatureAvailable(silentCheck=True))
        self._reloadModulesAct.triggered.connect(self._atReloadSelectedModules)
        self._reloadModulesAct.setIcon(self._reimportAndRunIcon)

        self._contextMenu.addAction(self._runAllAct)
        self._contextMenu.addAction(self._runSelectedAct)
        self._contextMenu.addSeparator()

        self._contextMenu.addAction(self._runSetupOnlyAct)
        self._contextMenu.addAction(self._runWithoutTearDownAct)
        self._contextMenu.addSeparator()

        self._contextMenu.addAction(self._copyPathAct)
        self._contextMenu.addAction(self._goToCodeAct)

        self._contextMenu.addSeparator()
        self._contextMenu.addAction(self._reloadModulesAct)

        self._logLevelMenu = QtWidgets.QMenu("Set Logging Level")
        self._contextMenu.addMenu(self._logLevelMenu)
        self._logLevelActionGrp = QtWidgets.QActionGroup(self._logLevelMenu)
        self._logLevelActionGrp.setExclusive(True)

    def _atReloadSelectedModules(self):
        for moduleItem in self.iterSelectedItemsOfCategories(
            constants.ITEM_CATEGORY_MODULE
        ):
            dotPath = self.testIdOfItem(moduleItem)
            importutils.reimportByModulePath(dotPath)

    def _atRunSetupOnly(self):
        testId = self.firstSelectedTestCaseId()
        if testId:
            self.runSetupOnly.emit(testId)

    def _atRunWithoutTearDown(self):
        testId = self.firstSelectedTestCaseId()
        if testId:
            self.runWithoutTearDown.emit(testId)

    def _atRunSelected(self):
        testIds = self.selectedTestIds(decomposePackageIfNecessary=True)
        if testIds:
            self.runTests.emit(testIds)

    def _atRunAll(self):
        self.runAllTest.emit()

    def _atGoToCode(self):
        firstSelectedModulePath = self._firstSelectedModulePath()
        sourceFile, line = pathutils.sourcePathAndLineFromModulePath(
            firstSelectedModulePath
        )
        if sourceFile:
            self._codeVisitor.goTo(sourceFile, line)

    def _setSelectedItemsLevel(self):
        level = variantToPyValue(self.sender().data())
        loggingutils.setLoggingLevel(level, *self.selectedTestIds())

    def contextMenuEvent(self, event):
        gotTests = self.hasTests()
        hasSelection = False
        hasSelectedTestOrCase = False
        hasModuleSelection = False
        if gotTests:
            hasSelection = self.hasSelectedTests(hasSelectedTestOrCase=False)
            hasSelectedTestOrCase = self.hasSelectedTests(hasSelectedTestOrCase=True)
            hasModuleSelection = self.hasSelectionOfCategories(
                constants.ITEM_CATEGORY_MODULE
            )

        self._runSelectedAct.setEnabled(hasSelection)
        self._runSetupOnlyAct.setEnabled(hasSelectedTestOrCase)
        self._runWithoutTearDownAct.setEnabled(hasSelectedTestOrCase)
        self._copyPathAct.setEnabled(hasSelection)
        self._goToCodeAct.setEnabled(hasSelection)
        self._reloadModulesAct.setEnabled(hasModuleSelection)

        self._logLevelMenu.clear()
        levels = loggingutils.loggingLevels(*self.selectedTestIds())
        currentLevel = levels[0] if len(levels) == 1 else None
        for level in loggingutils.allLoggingLevel():
            act = self._logLevelMenu.addAction(logging.getLevelName(level))
            act.setCheckable(True)
            act.setChecked(level == currentLevel)
            self._logLevelActionGrp.addAction(act)
            self._logLevelMenu.addAction(act)
            act.setData(level)
            act.triggered.connect(self._setSelectedItemsLevel)

        self._contextMenu.exec_(event.globalPos())



import unittest
import os

from iutest.core import iconutils
from iutest.core import constants


class IconUtilsTestCase(unittest.TestCase):
    def setUp(self):
        self._iconFile = "testCase.svg"

    def test_iconPath(self):
        iconPath = iconutils.iconPath(self._iconFile)
        self.assertTrue(iconPath)
        self.assertTrue(os.path.isfile(iconPath))

    def test_iconPathSet(self):
        iconPaths = iconutils.iconPathSet(self._iconFile, constants.TEST_ICON_SUFFIXES)
        self.assertTrue(iconPaths)
        for iconPath in iconPaths:
            self.assertTrue(iconPath)
            self.assertTrue(os.path.isfile(iconPath))



import unittest

from iutest.core import gotocode


class GoToCodeTestCase(unittest.TestCase):
    def setUp(self):
        self._path = "/path/to/source.py"
        self._line = 21

    def test_saveRestore(self):
        data = {
            'code --goto "$file":$line':'code --goto "/path/to/source.py":21',
            'notepad "$file"':'notepad "/path/to/source.py"',
            'vim +$line "$file"':'vim +21 "/path/to/source.py"',
        }
        for template, expected in data.items():
            self.assertEqual(
                gotocode.CodeLineVisitor._goToCmd(template, self._path, self._line),
                expected
            )



import unittest
import os
from iutest.core import pathutils


class PathUtilsTestCase(unittest.TestCase):
    def test_isPath(self):
        data = (
            ("/path/to/startDir", True),
            ("E:\\path\\to/startDir", True),
            ("module", False),
            ("package1.package2.module", False),
        )
        for _input, expectedResult in data:
            self.assertEqual(pathutils.isPath(_input), expectedResult)

    def test_iutestRootDir(self):
        rootDir = pathutils.iutestRootDir()
        self.assertTrue(os.path.isdir(rootDir))
        self.assertTrue("iutest", os.listdir(rootDir))
        self.assertTrue(os.path.isfile(os.path.join(rootDir, "iutest", "__init__.py")))

    def test_iutestPackageDir(self):
        rootDir = pathutils.iutestPackageDir()
        self.assertTrue(os.path.isdir(rootDir))
        self.assertTrue(os.path.isfile(os.path.join(rootDir, "__init__.py")))

    def test_objectFromDotPath(self):
        from iutest import core as corePackage
        from iutest.core import testmanager as testmanagerMod
        data = {
            "iutest.core":corePackage,
            "iutest.core.testmanager":testmanagerMod,
            "iutest.core.testmanager.TestManager":testmanagerMod.TestManager,
            "iutest.core.testmanager.TestManager.setRunnerMode":testmanagerMod.TestManager.setRunnerMode,
            "iutest.core.testmanager.logger":testmanagerMod.logger,
        }
        for dotPath, expectedObj in data.items():
            self.assertEqual(pathutils.objectFromDotPath(dotPath), expectedObj)

    def test_sourceFileAndLineFromObject(self):
        from iutest import core as corePackage
        from iutest.core import testmanager as testmanagerMod
        data = [
            (corePackage, False),
            (testmanagerMod, False),
            (testmanagerMod.TestManager, True), 
            (testmanagerMod.TestManager.setRunnerMode, True),
        ]
        for obj, notLine0 in data:
            srcFile, line = pathutils.sourceFileAndLineFromObject(obj)
            self.assertTrue(srcFile)
            self.assertTrue(os.path.isfile(srcFile))
            self.assertEqual(bool(line), notLine0)



import argparse
import logging
import sys
import os

logger = logging.getLogger(__name__)


def runUi(modulePathOrDir=None, topDir=None, exit_=False):
    """Load the IUTest UI

    Args:
        modulePathOrDir (str): The directory or the module path to search for tests.
        topDir (str): The top directory that need to be put in sys.path in order for the tests work.
        exit (bool): Whether we exit python console after the IUTest window closed.
    """
    from iutest import qt as _qt
    if not _qt.checkQtAvailability():
        return

    from iutest.ui import iutestwindow
    with _qt.ApplicationContext(exit_=exit_) as ctx:
        if ctx.isStandalone:
            logging.basicConfig()
        try:
            manager = iutestwindow.IUTestWindow(
                startDirOrModule=modulePathOrDir, topDir=topDir
            )
            manager.show()
        except Exception:
            ctx.setHasError(True)
            logger.exception("Error loading IUTest window.")


def runTests(runnerName, *testModulePathsOrDir, **arguments):
    """Run the tests without UI

    Args:
        runnerName (str): The runner name, e.g. 'nose2' or 'pytest'
        testModulePathsOrDir (tuple): List of python module paths or a single directory contains test modules.
        miscArguments (dict): Typical supported arguments are:
            topDir (str): The dir contains the python modules that the tests need for running.
            stopOnError (bool): Stop the tests running on the first error/failure.
    """
    from iutest.core import testmanager
    from iutest.core.runners import runnerconstants
    dirs = [s for s in testModulePathsOrDir if os.path.isdir(s)]
    if len(dirs) > 1:
        logger.error("Please only input at most one test root dir once at a time.")
        return

    testRootDir =  dirs[0] if dirs else None
    topDir  = arguments.get("topDir", None)
    stopOnError = arguments.get("stopOnError", False)
    manager = testmanager.TestManager(ui=None, startDirOrModule=testRootDir, topDir=topDir)
    manager.setRunnerMode(runnerconstants.runnerModeFromName(runnerName))
    manager.setStopOnError(stopOnError)
    if testRootDir:
        manager.runAllTests()
    else:
        manager.runTests(*testModulePathsOrDir)


def main():
    from iutest import _version
    from iutest.core.runners import runnerconstants

    parser = argparse.ArgumentParser(description="IUTest")
    parser.add_argument(
        "-v", "--version", action="version", version=_version.__version__
    )

    parser.add_argument(
        "-u",
        "--ui",
        action="store_true",
        dest="ui",
        default=False,
        help="Run test interactively in IUTest UI",
    )

    parser.add_argument(
        "-e",
        "--stopOnError",
        action="store_true",
        dest="stopOnError",
        default=False,
        help="Stop test running once there is an test error or failure",
    )

    parser.add_argument(
        "-t",
        "--topDir",
        action="store",
        dest="topDir",
        help="Specify the top dir of the python module to import",
    )

    parser.add_argument(
        "-m",
        "--runner",
        action="store",
        dest="runner",
        help="The runner the runs the tests, e.g. 'nose2' or 'pytest'",
    )

    parser.add_argument(
        "-r",
        "--runTests",
        action="append",
        default=[],
        dest="testPathsOrDir",
        help="Specify the top dir of the python module to import",
    )

    # parse the arguments from standard input
    results = parser.parse_args()
    if len(sys.argv) < 2 or results.ui:
        testDir = results.testPathsOrDir[0] if results.testPathsOrDir else None
        runUi(testDir, topDir=results.topDir)
    else:
        if not results.testPathsOrDir:
            print(
                "You need to specify -r/--runTests with a directory or python module paths to run the tests."
            )
            return

        if not runnerconstants.isValidRunnerName(results.runner):
            print(
                "You need to specify a valid test runner, e.g. --runner 'nose2' or --runner 'pytest'"
            )
            return

        arguments = {
            "topDir" : results.topDir,
            "stopOnError" : results.stopOnError,
        }
        runTests(
            results.runner,
            *results.testPathsOrDir,
            **arguments
        )


if __name__ == "__main__":
    main()



import logging
import os
from iutest.core import appsettings
from iutest.core import constants
from iutest.qt import QtCore

logger = logging.getLogger(__name__)


class CodeLineVisitor(QtCore.QObject):
    _editorSetting = None

    @classmethod
    def initEditorSetting(cls):
        cls._editorSetting = appsettings.get().simpleConfigStrValue(
            constants.CONFIG_KEY_CODE_EDITOR, constants.CONFIG_KEY_CODE_EDITOR_DEFAULT
        )

    @classmethod
    def config(cls):
        if not cls._editorSetting:
            cls.initEditorSetting()
        return cls._editorSetting

    def __init__(self, parent=None):
        QtCore.QObject.__init__(self, parent=parent)
        self._process = QtCore.QProcess(self)

    @staticmethod
    def _goToCmd(template, filePath, lineNumber):
        cmd = template.replace(constants.CODE_FILE_VAR, filePath)
        return cmd.replace(constants.CODE_LINE_VAR, str(lineNumber))

    def goTo(self, filePath, lineNumber=0):
        if os.path.isfile(filePath):
            logger.warning("%s is not a valid file.", filePath)
            
        cmd = self._goToCmd(self.config(), filePath, lineNumber)
        logger.debug(cmd)
        self._process.start(cmd)



from iutest.core import constants
import logging
import os
from unittest import loader, suite
from unittest import main as runPyUnittest
from iutest.core import pathutils
from iutest.core.runners import base
from iutest.core.runners import runnerconstants
from iutest.plugins.pyunitextentions import pyunitwrappers

logger = logging.getLogger(__name__)


class PyUnitRunner(base.BaseTestRunner):
    _Icon = None
    _Runner = None
    _lastTests = {}
    _gotError = False

    _loadTestsFailure = "_FailedTest"
    @classmethod
    def isValid(cls):
        return True

    @classmethod
    def mode(cls):
        return runnerconstants.RUNNER_PYUNIT

    @classmethod
    def iconFileName(cls):
        return "pyunit.svg"

    def runTests(self, *testIds):
        self._runTests(constants.RUN_TEST_FULL, *testIds)
        
    def _runTests(self, partialMode=constants.RUN_TEST_FULL, *testIds):
        failfast = self._manager.stopOnError()
        argv = [""]
        argv.extend(testIds)
        pyunitwrappers.PyUnitTestResult.resetLastData()
        testRunner = pyunitwrappers.PyUnitTestRunnerWrapper(
            failfast=failfast,
            partialMode=partialMode
        )
        runPyUnittest(
            None, 
            argv=argv,
            testRunner=testRunner,
            exit=False, 
            failfast=failfast, 
        )
        testRunner.resetUiStreamResult()

    def runSingleTestPartially(self, testId, partialMode):
        self._runTests(partialMode, testId)

    def _collectAllPaths(self, tests):
        if isinstance(tests, suite.TestSuite):
            for t in tests:
                for p in self._collectAllPaths(t):
                    yield p
        else:
            if not self._isTestFailedToLoad(tests):
                self.__class__._gotError = True
                return

            self._lastTests[tests.id()] = tests
            yield tests.id()

    @classmethod
    def _resetLastTests(cls):
        cls._lastTests = {}

    def _isTestFailedToLoad(self, test):
        if test.__class__.__name__ == self._loadTestsFailure:
            modulename = test.id().split(self._loadTestsFailure)[-1][1:]
            raiser = getattr(test, modulename, None)
            if not raiser:
                return True
            try:
                raiser()
            except Exception:
                logger.exception("Unable to load tests from directory %s", modulename)
                return False

        return True

    def iterAllTestIds(self):
        self.__class__._gotError = False
        self._resetLastTests()
        startDirOrModule = self._manager.startDirOrModule()
        startModule = pathutils.objectFromDotPath(startDirOrModule, silent=True)
        topDir = self._manager.topDir()
        if not topDir or not os.path.isdir(topDir):
            if os.path.isdir(startDirOrModule):
                topDir = startDirOrModule
            elif startModule:
                topDir = os.path.dirname(startModule.__file__)
        
        try:
            tests = loader.defaultTestLoader.discover(startDirOrModule, top_level_dir=topDir)
            if not startModule:
                for p in self._collectAllPaths(tests):
                    yield p
            else:
                for p in self._collectAllPaths(tests):
                    if not p.startswith(startDirOrModule):
                        yield ".".join([startDirOrModule, p])
        except Exception:
            self.__class__._gotError = True
            logger.exception("Unable to load tests from %s", startModule)

    @classmethod
    def hasLastListerError(cls):
        return cls._gotError

    @classmethod
    def lastRunInfo(cls):
        return pyunitwrappers.PyUnitTestResult.lastRunInfo
    
    @classmethod
    def avoidRunTestsOnPackageLevel(self):
        """PyUnit is difficult to run test from package level without discovering it.
        """
        return True



import sys
import logging
import time

from iutest.core import uistream
from iutest.core import pathutils
from iutest.core import runinfo
from iutest.core import constants
from iutest.core import pyunitutils

logger = logging.getLogger(__name__)

class PyUnitUiMixin(object):
    lastRunInfo = runinfo.TestRunInfo()

    _originalStdOut = sys.stdout
    _originalStdErr = sys.stderr
    
    def __init__(self, stream):
        self.stream = stream
        self.logHandler = uistream.LogHandler()
        self.stdOutCapturer = uistream.StdOutCapturer(self._originalStdOut)
        self.stdErrCapturer = uistream.StdErrCapturer(self._originalStdErr)

    def _linkInfoFromTest(self, test):
        try:
            testMethodName = test.id().split(".")[-1]
            test = getattr(test, testMethodName)
            sourceFile, line = pathutils.sourceFileAndLineFromObject(test)
            if not sourceFile:  # os.path.isfile(sourceFile)
                return None

            return testMethodName, sourceFile, line
        except:
            logger.debug("Unable retrieve test information for quick navigation.")
        return None

    @classmethod
    def resetLastData(cls):
        cls.lastRunInfo.reset()

    def _callUiMethod(self, method, *args, **kwargs):
        if hasattr(self.stream, "callUiMethod"):
            self.stream.callUiMethod(method, *args, **kwargs)

    @classmethod
    def _recordLastFailedTestId(cls, testId):
        if not cls.lastRunInfo.failedTestId:
            cls.lastRunInfo.failedTestId = testId

    def _atOutcomeAvailable(self, testId, resultCode):
        if resultCode == constants.TEST_RESULT_ERROR:
            self.Cls.lastRunInfo.errorCount += 1
            self._recordLastFailedTestId(testId)

        elif resultCode == constants.TEST_RESULT_FAIL:
            self.Cls.lastRunInfo.failedCount += 1
            self._recordLastFailedTestId(testId)

        elif resultCode == constants.TEST_RESULT_EXPECTED_FAIL:
            self.Cls.lastRunInfo.expectedFailureCount += 1
            self._recordLastFailedTestId(testId)

        elif resultCode == constants.TEST_RESULT_UNEXPECTED_PASS:
            self.Cls.lastRunInfo.unexpectedSuccessCount += 1

        elif resultCode == constants.TEST_RESULT_SKIP:
            self.Cls.lastRunInfo.skipCount += 1

        elif resultCode == constants.TEST_RESULT_PASS:
            self.Cls.lastRunInfo.successCount += 1

        self._callUiMethod(
                "showResultOnItemByTestId", testId, resultCode
            )
        self.stdOutCapturer.stop()
        self.stdErrCapturer.stop()
        self.logHandler.stop()

    def _atStartTestRun(self):
        self._callUiMethod(
            "repaintUi"
        )  # To avoid double clicking to run single test will end up massive selection.
        self.Cls.lastRunInfo.failedTestId = None
        self.Cls.lastRunInfo._sessionStartTime = time.time()
        self.Cls.lastRunInfo._testStartTimes = {}
        self._callUiMethod("onTestRunningSessionStart")

    def _startLogProcessers(self):
        self.logHandler.start()
        self.stdOutCapturer.start()
        self.stdErrCapturer.start()

    def _atStartTest(self, test):
        self.Cls.lastRunInfo.runCount += 1
        originalTestId = test.id()
        _, testId = pyunitutils.parseParameterizedTestId(originalTestId)
        if testId not in self.Cls.lastRunInfo.runTestIds:
            testStartTime = time.time()
            self.Cls.lastRunInfo._testStartTimes[testId] = testStartTime
            self.Cls.lastRunInfo.runTestIds.append(testId)
            self._callUiMethod("onSingleTestStart", testId, testStartTime)

    def _atStopTest(self, test):
        originalTestId = test.id()
        _, testId = pyunitutils.parseParameterizedTestId(originalTestId)
        stopTime = time.time()
        testStartTime = self.Cls.lastRunInfo._testStartTimes.get(testId, self.Cls.lastRunInfo._sessionStartTime)
        self.Cls.lastRunInfo.singleTestRunTime = stopTime - testStartTime
        
        self._callUiMethod("onSingleTestStop", testId, stopTime)
        self._callUiMethod("repaintUi")

    def _atStopTestRun(self):
        self.Cls.lastRunInfo.sessionRunTime = time.time() - self.Cls.lastRunInfo._sessionStartTime
        self._callUiMethod("onAllTestsFinished")



from iutest.qt import QtCore, QtWidgets


def makeMainLayout(hostWidget=None):
    _lay = QtWidgets.QVBoxLayout(hostWidget)
    _lay.setContentsMargins(6, 6, 6, 6)
    _lay.setSpacing(3)
    return _lay


def makeMinorLayout(layoutClass):
    _layout = layoutClass()
    _layout.setSpacing(3)
    _layout.setContentsMargins(0, 0, 0, 0)
    return _layout


def makeMinorHorizontalLayout():
    return makeMinorLayout(QtWidgets.QHBoxLayout)


def makeMinorVerticalLayout():
    return makeMinorLayout(QtWidgets.QVBoxLayout)


def makeIconButton(icon, parent=None):
    btn = QtWidgets.QPushButton("", parent)
    btn.setIcon(icon)
    btn.setFlat(True)
    btn.setIconSize(QtCore.QSize(20, 20))
    btn.setFixedSize(QtCore.QSize(24, 24))
    return btn


def addSeparatorToLayout(layout, direction):
    sep = QtWidgets.QFrame()
    sep.setStyleSheet("color:black")
    policy = sep.sizePolicy()
    if direction == QtCore.Qt.Horizontal:
        sep.setFrameShape(sep.HLine)
        sep.setSizePolicy(policy.Minimum, policy.Preferred)
    else:
        sep.setFrameShape(sep.VLine)
        sep.setSizePolicy(policy.Preferred, policy.Minimum)

    sep.setLineWidth(1)
    layout.addWidget(sep)



import logging

logger = logging.getLogger(__name__)


def allLoggingLevel():
    return (
        logging.NOTSET,
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.FATAL,
    )


def loggingLevels(*dotPaths):
    return [logging.getLogger(dotPath).level for dotPath in dotPaths]


def setLoggingLevel(level, *dotPaths):
    levelName = logging.getLevelName(level)
    for dotPath in dotPaths:
        logging.getLogger(dotPath).setLevel(level)
        logging.info("Set logging level of %s to: %s", dotPath, levelName)



import logging
import os

'''
from iutest import dependencies
pytest = dependencies.PyTestWrapper.getModule()

logger = logging.Logger(__name__)

def _power(inputValue):
    return inputValue * inputValue


def test_simpleApproved():
    """ a doctest in a docstring
    >>> _power(5)
    25
    >>> _power(4)
    12
    """
    assert _power(5) == 25
    

def test_simpleFailed():
    assert _power(5) == 6


@pytest.fixture()
def _prepareSomthing(tmpdir_factory):
    return tmpdir_factory.mktemp("data").join("temp.txt")


def test_preparedTempFile(_prepareSomthing):
    print ("Prepared file:", _prepareSomthing)
    print (dir(_prepareSomthing))


def test_toSkip():
    print (_power(5))
    pytest.skip('This test just to be skipped')



# ---------------------------------------------

class TestSomething:
    def test_ok(self):
        x = "this"
        assert "h" in x

    def test_failed(self):
        x = "hello"
        assert hasattr(x, "check")

'''



import unittest
import os
import sys
import tempfile
import shutil

import iutest
from iutest import cli
from iutest.core import pathutils
from iutest.core.runners import runnerconstants

from iutest.tests.iutests import test_runnercommon as common


class CliTestCase(unittest.TestCase):
    def setUp(self):
        self._tempModuleName = "test_iutest_cli_temp"
        self._tempDir = None
        self._runner = runnerconstants.RUNNER_NAMES[runnerconstants.RUNNER_PYUNIT]
        common.initTestTargets(self)

    def tearDown(self):
        if self._tempModuleName in sys.modules:
            del sys.modules[self._tempModuleName]
        
        if self._tempDir and os.path.isdir(self._tempDir):
            shutil.rmtree(self._tempDir)
            if self._tempDir in sys.path:
                sys.path.remove(self._tempDir)

    def _buildTempTests(self):
        if not self._tempDir or not os.path.isdir(self._tempDir):
            self._tempDir = tempfile.mkdtemp()

        # common.__file__ could be pyc file, what we need is py file:
        srcFile = os.path.join(os.path.dirname(common.__file__), "test_runnercommon.py")
        if not os.path.isfile(srcFile):
            return

        with open(srcFile, "r") as f:
            content = f.read()
            
        content = content.replace("RunnerDummyTestCase", "TempRunnerDummyTestCase")
        dstFile = os.path.join(self._tempDir, "{}.py".format(self._tempModuleName))
        with open(dstFile, "w") as f:
            f.write(content)
        self.assertTrue(os.path.isfile(dstFile))

        initFile = os.path.join(self._tempDir, "__init__.py")
        with open(initFile, "w"):
            pass
        self.assertTrue(os.path.isfile(initFile))

        if self._tempDir not in sys.path:
            sys.path.append(self._tempDir)

    def test_cliRunTestsByModulePath(self):
        runAction = lambda : cli.runTests(self._runner, self._testId)
        common.checkRunTestsAfterAction(self, runAction)

    def test_cliRunTestsByDirs(self):
        self._buildTempTests()
        import test_iutest_cli_temp as tempcommon
        runAction = lambda : cli.runTests(self._runner, self._tempDir)
        tempcommon.checkRunTestsAfterAction(self, runAction)

    def test_iutestRunTestsByModulePath(self):
        runAction = lambda : iutest.runTests(self._runner, self._testId)
        common.checkRunTestsAfterAction(self, runAction)

    def test_iurestRunTestsByDirs(self):
        self._buildTempTests()
        import test_iutest_cli_temp as tempcommon
        runAction = lambda : iutest.runTests(self._runner, self._tempDir)
        tempcommon.checkRunTestsAfterAction(self, runAction)



import unittest
import logging
import nose2

from iutest.core import constants

logger = logging.getLogger(__name__)


class _NoTearDown(object):
    def __init__(self, test):
        self._test = test
        self._originalTearDown = test.tearDown

    def _dummyTearDown(self):
        logger.info("Skipped %s.tearDown().", self._test.__class__.__name__)

    def __enter__(self):
        self._test.tearDown = self._dummyTearDown

    def __exit__(self, *args, **kwargs):
        self._test.tearDown = self._originalTearDown


class PartialTestRunner(nose2.events.Plugin):
    """A nose2 plug to run parts of single test, like run only setup, setup+test, etc.
    """

    commandLineSwitch = (None, "partial-test", "Run partial test.")
    runMode = constants.RUN_TEST_SETUP_ONLY

    lastRunTest = None

    @classmethod
    def setRunMode(cls, mode):
        """Set run mode to either TEST_SETUP_ONLY or TEST_NO_TEAR_DOWN
        """
        cls.runMode = min(
            constants.RUN_TEST_FULL,
            max(constants.RUN_TEST_SETUP_ONLY, int(mode)),
        )

    def registerInSubprocess(self, event):
        event.pluginClasses.append(self.__class__)

    def startTestRun(self, event):
        PartialTestRunner.lastRunTest = None
        event.executeTests = self.runPartialTest

    def startSubprocess(self, event):
        PartialTestRunner.lastRunTest = None
        event.executeTests = self.runPartialTest
        event.StopTestRunEvent = None

    def _runPartialTest(self, test):
        if self.runMode == constants.RUN_TEST_SETUP_ONLY:
            test.setUp()
            logger.info("Run %s.setUp() only.", test.__class__.__name__)
        elif self.runMode == constants.RUN_TEST_NO_TEAR_DOWN:
            with _NoTearDown(test):
                test.run(self.session.testResult)
        else:
            test.run(self.session.testResult)

        PartialTestRunner.lastRunTest = test

    def stopTestRun(self, event):
        if self.lastRunTest:
            event.result.shouldStop = True

    def runPartialTest(self, suite, result):
        """Collect tests, but don't run them"""
        for test in suite:
            if isinstance(test, unittest.BaseTestSuite):
                self.runPartialTest(test, result)
                continue

            self._runPartialTest(test)
            return



import unittest
import os
import logging

from iutest.core import loggingutils
from iutest.core import constants


class LoggingUtilsTestCase(unittest.TestCase):
    def test_loggingLevelEdit(self):
        modules = [
            "iutest.core",
            "iutest.core.loggingutils",
            "iutest.core.constants",
        ]
        allLevels = loggingutils.allLoggingLevel()
        self.assertTrue(allLevels)
        for level in allLevels:
            expectedLevels = [level]*len(modules)
            loggingutils.setLoggingLevel(level, *modules)
            self.assertEqual(list(loggingutils.loggingLevels(*modules)), expectedLevels)



from iutest.core import importutils


def runUi(modulePathOrDir=None, topDir=None, exit_=False):
    """Load the IUTest UI

    Args:
        modulePathOrDir (str): The directory or the module path to search for tests.
        topDir (str): The top directory that need to be put in sys.path in order for the tests work.
        exit (bool): Whether we exit python console after the IUTest window closed.
    """
    from iutest import cli
    cli.runUi(modulePathOrDir=modulePathOrDir, topDir=topDir, exit_=exit_)


def runTests(runnerName, *testModulePathsOrDir, **arguments):
    """Run the tests without UI

    Args:
        runnerName (str): The runner name, e.g. 'nose2' or 'pytest'
        testModulePathsOrDir (tuple): List of python module paths or a single directory contains test modules.
        miscArguments (dict): Typical supported arguments are:
            topDir (str): The dir contains the python modules that the tests need for running.
            stopOnError (bool): Stop the tests running on the first error/failure.
    """
    from iutest import cli
    cli.runTests(runnerName, *testModulePathsOrDir, **arguments)


__all__ = ["importutils", "runUi", "runTests"]



import logging

from iutest.core import constants
from iutest.core import appsettings
from iutest.core import gotocode
from iutest.qt import QtCore, QtWidgets
from iutest.ui import uiutils

logger = logging.getLogger(__name__)


class ConfigWindow(QtWidgets.QDialog):
    """As there are not much to config and there won't be many in forseeable future, we use simple, 
    static way to generate these widgets.
    """

    _instance = None

    def __init__(self, parent=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.setWindowTitle("{} Preference".format(constants.APP_NAME))

        self._mainLayout = uiutils.makeMainLayout(self)
        self._formLayout = uiutils.makeMinorLayout(QtWidgets.QFormLayout)
        self._mainLayout.addLayout(self._formLayout)
        self._formLayout.setLabelAlignment(QtCore.Qt.AlignRight)

        # Code viewing config
        self._codeEditorLE = QtWidgets.QLineEdit(self)
        editorSetting = gotocode.CodeLineVisitor.config()
        self._codeEditorLE.setText(editorSetting)
        self._codeEditorLE.editingFinished.connect(self._onCodeEditorEditFinished)
        self._codeEditorLE.setPlaceholderText("Example: pargram $file -argToLine $line")
        self._codeEditorLE.setToolTip(
            "Input the command to jump to the code at the line. \n"
            "$file is the placeholder for file path, $line is the line number."
        )
        _annoText = QtWidgets.QLabel(
            "Default: {}".format(constants.CONFIG_KEY_CODE_EDITOR_DEFAULT)
        )
        _annoText.setEnabled(False)

        self._formLayout.addRow("Go To Code Line", self._codeEditorLE)
        self._formLayout.addRow("", _annoText)

        self.setMinimumWidth(400)
        self.setMinimumHeight(100)

        self.destroyed.connect(self._onDialogDeleted)

    def _onCodeEditorEditFinished(self):
        txt = str(self._codeEditorLE.text()).strip()
        appsettings.get().saveSimpleConfig(constants.CONFIG_KEY_CODE_EDITOR, txt)
        gotocode.CodeLineVisitor.initEditorSetting()

    @classmethod
    def _onDialogDeleted(cls, *_):
        cls._instance = None

    @classmethod
    def show(cls, parent=None):
        if not cls._instance:
            cls._instance = cls(parent)
        else:
            cls._instance.setParent(parent)
        cls._instance.exec_()



import logging

from iutest.core import uistream
from iutest.core import constants
from iutest.core import runinfo
from iutest.plugins.pyunitextentions import pyunitcommon

from nose2.plugins import result as resultPlugin  # This has to be imported this way.
from nose2 import result, events  # This has to be imported this way.

logger = logging.getLogger(__name__)


class UiHooksPlugin(resultPlugin.ResultReporter, pyunitcommon.PyUnitUiMixin):
    """A nose2 plug to capture the logs for ui.
    """
    lastRunInfo = runinfo.TestRunInfo()

    def __init__(self):
        resultPlugin.ResultReporter.__init__(self)
        pyunitcommon.PyUnitUiMixin.__init__(self, uistream.UiStream())
        self.Cls = self.__class__

    @classmethod
    def _mapTestResultCode(cls, resultCode, expected):
        """Translate nose2.result result constants to iutest.core.constants.TEST_RESULT_* constants
        """
        if resultCode == result.ERROR:
            return constants.TEST_RESULT_ERROR
        
        if resultCode == result.FAIL:
            if expected:
                return constants.TEST_RESULT_EXPECTED_FAIL
            return constants.TEST_RESULT_FAIL

        if resultCode == result.SKIP:
            return constants.TEST_RESULT_SKIP

        if resultCode == result.PASS:
            if not expected:
                return constants.TEST_RESULT_UNEXPECTED_PASS
            return constants.TEST_RESULT_PASS

        return constants.TEST_RESULT_NONE

    def startTestRun(self, event):
        self._atStartTestRun()

    def startTest(self, event):
        self._atStartTest(event.test)
        with self.stream.linkInfoCtx(self._linkInfoFromTest(event.test)):            
            resultPlugin.ResultReporter.startTest(self, event)

        self._startLogProcessers()

    def stopTest(self, event):
        self._atStopTest(event.test)

    def testOutcome(self, event):
        testId = event.test.id()
        resultCode = self._mapTestResultCode(event.outcome, event.expected)
        self._atOutcomeAvailable(testId, resultCode)      

        with self.stream.resultCtx(resultCode):
            resultPlugin.ResultReporter.testOutcome(self, event)

    def stopTestRun(self, event):
        self._atStopTestRun()

    def afterTestRun(self, event):
        evt = events.ReportSummaryEvent(event, self.stream, self.reportCategories)

        self.stream.setLinkInfo()
        self.stream.setProcessStackTraceLink(False)

        if evt.stopTestEvent.result.wasSuccessful():
            with self.stream.resultCtx(self._mapTestResultCode(result.PASS, True)):
                resultPlugin.ResultReporter.afterTestRun(self, event)
        else:
            cats = evt.reportCategories
            errors = cats.get("errors", [])
            failures = cats.get("failures", [])
            for err in errors:
                self._reportErrorSummary("ERROR", err)

            for fail in failures:
                self._reportErrorSummary("FAIL", fail)

            for flavour, events_ in cats.items():
                if flavour in self.dontReport:
                    continue

                for ev in events_:
                    self._reportErrorSummary(flavour.upper(), ev)

            with self.stream.resultCtx(self._mapTestResultCode(result.FAIL, False)):
                self._printSummary(evt)

    def _reportErrorSummary(self, flavour, err):
        self.stream.setProcessStackTraceLink(False)

        desc = self._getDescription(err.test, errorList=True)
        with self.stream.resultCtx(self._mapTestResultCode(result.FAIL, False)):
            self.stream.writeln(self.separator1)

        with self.stream.linkInfoCtx(self._linkInfoFromTest(err.test)):
            self.stream.writeln("%s: %s" % (flavour, desc))

        # Now the stack trace, etc:
        self.stream.writeln(self.separator2)
        errDetail = self._getOutcomeDetail(err)
        if errDetail:
            with self.stream.processStackTraceLinkCtx():
                self.stream.writeln(errDetail)



import unittest
import os
from unittest.suite import TestSuite

from iutest.core import constants
from iutest.core import pathutils
from iutest.core import pyunitutils
from iutest.core import testmanager

from iutest.core.runners import runnerconstants

class RunnerDummyTestCase(unittest.TestCase):
    _setupExecuted = False
    _testExecuted = False
    _tearDownExecuted = False

    def setUp(self):
        self.__class__._setupExecuted = True
        self.__class__._testExecuted = False
        self.__class__._tearDownExecuted = False

    def test_dummy(self):
        self.__class__._testExecuted = True
        self.__class__._tearDownExecuted = False

    def tearDown(self):
        self.__class__._tearDownExecuted = True

    @classmethod
    def checkSetUpRun(cls, testSuite, expectRun):
        testSuite.assertEqual(cls._setupExecuted, expectRun)

    @classmethod
    def checkTestRun(cls, testSuite, expectRun):
        testSuite.assertEqual(cls._testExecuted, expectRun)

    @classmethod
    def checkTearDownRun(cls, testSuite, expectRun):
        testSuite.assertEqual(cls._tearDownExecuted, expectRun)

    @classmethod
    def resetStates(cls, testSuite):
        cls._setupExecuted = False
        cls._testExecuted = False
        cls._tearDownExecuted = False
        cls.checkSetUpRun(testSuite, False)
        cls.checkTestRun(testSuite, False)
        cls.checkTearDownRun(testSuite, False)


def initTestTargets(testSuite):
    testSuite._testId = __name__
    testSuite._modulePath = "iutest.tests"


def setUpTest(testSuite, runnerMode):
    testSuite._manager = testmanager.TestManager(None, None)
    testSuite._manager.setRunnerMode(runnerMode)
    testSuite.assertEqual(testSuite._manager.getRunner().mode(), runnerMode)
    initTestTargets(testSuite)

def checkListedTests(testSuite, tests):
    testSuite.assertTrue(tests)
    hasThisTest = False
    prefix = "iutest.tests.iutests.{}.{}.test_".format(
        os.path.splitext(os.path.basename(__file__))[0],
        RunnerDummyTestCase.__name__, 
    )
    for tid in tests:
        if tid.startswith(prefix):
            hasThisTest = True
    testSuite.assertTrue(hasThisTest)


def checkParseParameterizedTestId(testSuite):
    data = {
        "test_mod.TestSuite.test_case":(False, "test_mod.TestSuite.test_case"),
        "test_mod.TestSuite.test_case:34":(True, "test_mod.TestSuite.test_case"),
    }
    for testId, expected in data.items():
        testSuite.assertEqual(pyunitutils.parseParameterizedTestId(testId), expected)


def checkListByDir(testSuite):
    startDir = os.path.join(pathutils.iutestPackageDir(), "tests")
    topDir = pathutils.iutestRootDir()
    testSuite._manager.setDirs(startDir, topDir)

    tests = list(testSuite._manager.iterAllTestIds())
    checkListedTests(testSuite, tests)


def checkListByModulePath(testSuite):
    testSuite._manager.setStartDirOrModule(testSuite._modulePath)
    
    tests = list(testSuite._manager.iterAllTestIds())
    checkListedTests(testSuite, tests)


def checkRunTestsAfterAction(testSuite, action):
    RunnerDummyTestCase.resetStates(testSuite)
    action()
    RunnerDummyTestCase.checkSetUpRun(testSuite, True)
    RunnerDummyTestCase.checkTestRun(testSuite, True)
    RunnerDummyTestCase.checkTearDownRun(testSuite, True)
    print("The test on {} did run.".format(RunnerDummyTestCase.__name__))


def checkRunTests(testSuite):
    testSuite._manager.setStartDirOrModule(testSuite._modulePath)
    action = lambda : testSuite._manager.runTests(testSuite._testId)
    checkRunTestsAfterAction(testSuite, action)


def checkPartialRun(testSuite):
    RunnerDummyTestCase.resetStates(testSuite)
    testSuite._manager.setStartDirOrModule(testSuite._modulePath)

    testSuite._manager.runSingleTestPartially(testSuite._testId, constants.RUN_TEST_SETUP_ONLY)
    RunnerDummyTestCase.checkSetUpRun(testSuite, True)
    RunnerDummyTestCase.checkTestRun(testSuite, False)
    RunnerDummyTestCase.checkTearDownRun(testSuite, False)
    RunnerDummyTestCase.resetStates(testSuite)

    testSuite._manager.runSingleTestPartially(testSuite._testId, constants.RUN_TEST_NO_TEAR_DOWN)
    RunnerDummyTestCase.checkSetUpRun(testSuite, True)
    RunnerDummyTestCase.checkTestRun(testSuite, True)
    RunnerDummyTestCase.checkTearDownRun(testSuite, False)
    RunnerDummyTestCase.resetStates(testSuite)

    testSuite._manager.runSingleTestPartially(testSuite._testId, constants.RUN_TEST_FULL)
    RunnerDummyTestCase.checkSetUpRun(testSuite, True)
    RunnerDummyTestCase.checkTestRun(testSuite, True)
    RunnerDummyTestCase.checkTearDownRun(testSuite, True)


def checkLastRunInfo(testSuite):
    testSuite._manager.setStartDirOrModule(testSuite._modulePath)
    testSuite._manager.runTests(testSuite._testId)
    lastRunInfo = testSuite._manager.lastRunInfo()
    testSuite.assertTrue(lastRunInfo)
    testSuite.assertTrue(lastRunInfo.runTestIds[0].startswith(testSuite._testId))
    testSuite.assertEqual(lastRunInfo.failedTestId, None)
    testSuite.assertTrue(lastRunInfo.runCount)
    testSuite.assertTrue(lastRunInfo.successCount)



import unittest
import os
from iutest.tests.iutests import test_runnercommon as common
from iutest import dependencies
from iutest.core import pathutils
from iutest.core import pyunitutils
from iutest.core.runners import runnerconstants


@unittest.skipUnless(dependencies.Nose2Wrapper.get().isValid(), "It is nose2 only test")
class Nose2RunnerTestCase(unittest.TestCase):
    def setUp(self):
        common.setUpTest(self, runnerconstants.RUNNER_NOSE2)

    def test_parseParameterizedTestId(self):
        common.checkParseParameterizedTestId(self)

    def test_listByDir(self):
        common.checkListByDir(self)

    def test_listByModulePath(self):
        common.checkListByModulePath(self)

    def test_runTests(self):
        common.checkRunTests(self)

    def test_partialRun(self):
        common.checkPartialRun(self)

    def test_lastRunInfo(self):
        common.checkLastRunInfo(self)



from iutest import qt as _qt

_APP_MODE_STANDALONE = "standalone"
_APP_MODE_MAYA = "maya"

__appMode = None


def appMode():
    global __appMode
    if __appMode is not None:
        return __appMode

    __appMode = _APP_MODE_STANDALONE
    try:
        import maya
    except:
        pass
    else:
        __appMode = _APP_MODE_MAYA

    return __appMode


def isInsideMaya():
    return appMode() == _APP_MODE_MAYA


def isStandalone():
    return appMode() == _APP_MODE_STANDALONE


def _dccModule():
    if isStandalone():
        from iutest.dcc import standalone
        return standalone
    elif isInsideMaya():
        from iutest.dcc import dccmaya
        return dccmaya


module = _dccModule()
findParentWindow = module.findParentWindow



from iutest.qt import QtWidgets, Signal


class RootPathEdit(QtWidgets.QLineEdit):
    rootPathChanged = Signal(str, str)

    def __init__(self, parent=None):
        QtWidgets.QLineEdit.__init__(self, parent)
        self.editingFinished.connect(self.onEditFinished)
        self._initPath = None

    def setInitialPath(self, path):
        self._initPath = path

    def setText(self, txt):
        QtWidgets.QLineEdit.setText(self, txt)
        self.setInitialPath(txt)

    def onEditFinished(self):
        txt = str(self.text())
        if txt == self._initPath:
            return

        self.setInitialPath(txt)
        self.rootPathChanged.emit(txt, "")



import logging
from iutest.core.runners import base
from iutest.core.runners import runnerconstants

logger = logging.getLogger(__name__)


class PyTestTestRunner(base.BaseTestRunner):
    # To-Do: Implement this class
    @classmethod
    def mode(cls):
        return runnerconstants.RUNNER_PYTEST

    @classmethod
    def isValid(cls):
        # To-Do: Enable it when it is ready.
        # return dependencies.PyTestWrapper.get().isValid()
        return False

    @classmethod
    def iconFileName(cls):
        return "pytest.svg"

    def runTests(self, *testIds):
        pass

    def runSingleTestPartially(self, testId, partialMode):
        """Run partial steps of test, like running setUp only, or setUp and test but without teardown.
        Args:
            testId (str): id of test, the python module.
            partialMode (int): the test run mode, available values are:
                constants.RUN_TEST_SETUP_ONLY | constants.RUN_TEST_NO_TEAR_DOWN
        """
        pass

    def iterAllTestIds(self):
        return iter([])




def parseParameterizedTestId(testId):
    isParameterized = ":" in testId
    return (
        (isParameterized, testId.split(":")[0])
        if isParameterized
        else (isParameterized, testId)
    )


import logging

from iutest.core import pathutils
from iutest.core.runners import runnerconstants
from iutest.core.runners import registry

logger = logging.getLogger(__name__)


class TestManager(object):
    def __init__(self, ui, startDirOrModule, topDir=None):
        self._startDirOrModule = ""
        self._topDir = ""
        self._stopOnError = False
        self._ui = ui
        self._runnerMode = runnerconstants.RUNNER_PYUNIT
        self._runners = {}
        self.setStartDirOrModule(startDirOrModule)
        self.setTopDir(topDir)
        self._verbosity = 1

    def setVerbosity(self, verbosity):
        self._verbosity = verbosity

    def verbosity(self):
        return self._verbosity

    def setRunnerMode(self, runnerMode):
        """Switch to different runner for listing or running the tests.
        Args:
            runnerMode (int): One of RUNNER_* int constant defined in registry.
        """
        self._initialiseRunner(runnerMode)
        self._runnerMode = runnerMode
        self.getRunner().check()

    def iterAllRunners(self):
        for runnerMode in registry.getRunnerModes():
            runner = self.getRunnerByMode(runnerMode)
            if runner:
                yield runner

    def getFeasibleRunnerMode(self):
        for runnerMode in registry.getRunnerModes():
            runner = self.getRunnerByMode(runnerMode)
            if runner.isValid():
                return runnerMode

        return runnerconstants.RUNNER_PYUNIT

    def setStartDirOrModule(self, startDirOrModule):
        self._startDirOrModule = startDirOrModule or ""
        if self._startDirOrModule:
            if not self._topDir or not self._startDirOrModule.startswith(self._topDir):
                self._topDir = self._startDirOrModule

    def startDirOrModule(self):
        return self._startDirOrModule

    def setTopDir(self, topDir):
        self._topDir = topDir or ""
        if self._topDir:
            if not self._startDirOrModule or not self._startDirOrModule.startswith(
                self._topDir
            ):
                self._startDirOrModule = self._topDir

    def topDir(self):
        return self._topDir

    def setDirs(self, startDirOrModule, topDir=None):
        startDirOrModule = str(startDirOrModule) if startDirOrModule else ""
        self.setStartDirOrModule(startDirOrModule)
        topDir = topDir or self._topDir
        topDir = str(topDir) if topDir else "" 
        if startDirOrModule:
            if topDir:
                if not self._topDir or not topDir.startswith(self._topDir):
                    self._topDir = topDir
            else:
                if not self._topDir or not startDirOrModule.startswith(self._topDir):
                    self._topDir = startDirOrModule

    def setStopOnError(self, stop):
        self._stopOnError = stop

    def stopOnError(self):
        return self._stopOnError

    def ui(self):
        return self._ui

    def _initialiseRunner(self, runnerMode):
        if runnerMode in self._runners:
            return True

        if runnerMode not in registry.RUNNER_REGISTRY:
            logger.error("The runner %s is invalid.", runnerMode)
            return False

        runnerModulePath = registry.RUNNER_REGISTRY[runnerMode]
        runnerClass = pathutils.objectFromDotPath(runnerModulePath)
        self._runners[runnerMode] = runnerClass(self)
        return True

    def getRunnerByMode(self, runnerMode):
        if not self._initialiseRunner(runnerMode):
            logger.error("Unable to initialise the runner of mode %s", runnerMode)
            return None

        return self._runners[runnerMode]

    def getRunner(self):
        return self.getRunnerByMode(self._runnerMode)

    def runTests(self, *tests):
        self.getRunner().runTests(*tests)

    def iterAllTestIds(self):
        for testId in self.getRunner().iterAllTestIds():
            yield testId

    def runAllTests(self):
        tests = list(self.iterAllTestIds())
        if not tests:
            logger.warning("No tests found to run.")
            return

        self.runTests(*tests)

    def runSingleTestPartially(self, testId, partialMode):
        """Run partial steps of test, like running setUp only, or setUp and test but without teardown.
        Args:
            testId (str): id of test, the python module.
            partialMode (int): the test run mode, available values are:
                constants.RUN_TEST_SETUP_ONLY | constants.RUN_TEST_NO_TEAR_DOWN
        """
        self.getRunner().runSingleTestPartially(testId, partialMode)

    def hasLastListerError(self):
        return self.getRunner().hasLastListerError()

    def lastRunTestIds(self):
        return self.getRunner().lastRunTestIds()

    def lastRunInfo(self):
        return self.getRunner().lastRunInfo()

    def parseParameterizedTestId(self, testId):
        return self.getRunner().parseParameterizedTestId(testId)

    def lastFailedTestId(self):
        return self.getRunner().lastFailedTestId()

    def avoidRunTestsOnPackageLevel(self):
        return self.getRunner().avoidRunTestsOnPackageLevel()


import logging
from iutest import qt as _qt
from iutest.core import pyunitutils
from iutest.core import iconutils
from iutest.core import runinfo
from iutest.core.runners import runnerconstants

logger = logging.getLogger(__name__)


class BaseTestRunner(object):
    def __init__(self, manager):
        self._manager = manager

    @classmethod
    def isValid(cls):
        return False

    @classmethod
    def iconFileName(cls):
        cls._raiseNotImplementedError()

    @classmethod
    def icon(cls):
        return _qt.iconFromPath(iconutils.iconPath(cls.iconFileName()))

    @classmethod
    def _issueNotInstalledError(cls):
        logger.warning(
            "The test runner %s is not installed, consider installing it by: `pip install %s`",
            cls.name(), cls.name(),
        )

    @classmethod
    def check(cls):
        if cls.isValid():
            logger.debug("Switch to test runner %s.", cls.name())
        else:
            logger.warning(
                "The test runner %s is unavailable, you need to install it first or switch to other runners.",
                cls.name(),
            )

    @classmethod
    def mode(cls):
        cls._raiseNotImplementedError()

    @classmethod
    def name(cls):
        return runnerconstants.RUNNER_NAMES[cls.mode()]

    @classmethod
    def _raiseNotImplementedError(cls):
        raise NotImplementedError(
            "You should not use {} directly.".format(cls.__name__)
        )

    def runTests(self, *testIds):
        self._raiseNotImplementedError()

    def runSingleTestPartially(self, testId, partialMode):
        """Run partial steps of test, like running setUp only, or setUp and test but without teardown.
        Args:
            testId (str): id of test, the python module.
            partialMode (int): the test run mode, available values are:
                constants.RUN_TEST_SETUP_ONLY | constants.RUN_TEST_NO_TEAR_DOWN
        """
        self._raiseNotImplementedError()

    def iterAllTestIds(self, startDir, topDir):
        self._raiseNotImplementedError()

    @classmethod
    def hasLastListerError(cls):
        pass

    @classmethod
    def lastRunInfo(cls):
        return runinfo.TestRunInfo()

    @classmethod
    def lastRunTestIds(cls):
        return cls.lastRunInfo().runTestIds

    @classmethod
    def lastFailedTestId(cls):
        return cls.lastRunInfo().failedTestId

    @classmethod
    def parseParameterizedTestId(cls, testId):
        return pyunitutils.parseParameterizedTestId(testId)
    
    @classmethod
    def avoidRunTestsOnPackageLevel(self):
        """A stupid interface to tell the test manager to split these package level test id first.
        """
        return False


from iutest.qt import QtWidgets, QtCore
from iutest.ui import uiconstants


class StatusLabel(QtWidgets.QLabel):
    def __init__(self, parent):
        QtWidgets.QLabel.__init__(self, parent)
        self.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        self._testManager = None

    def setTestManager(self, manager):
        self._testManager = manager

    def updateReport(self):
        runInfo = self._testManager.lastRunInfo()
        testLbl = "tests" if runInfo.runCount != 1 else "test"
        msgs = [
            "%s %s run in %.3f sec"
            % (runInfo.runCount, testLbl, runInfo.sessionRunTime)
        ]
        if runInfo.successCount == runInfo.runCount:
            msgs.append(
                "<font color=%s>ALL GOOD :)</font>"
                % uiconstants.LOG_COLOR_SUCCESS.name()
            )
        else:
            if runInfo.failedCount:
                msgs.append(
                    "<font color=%s>%s failed</font>"
                    % (uiconstants.LOG_COLOR_FAILED.name(), runInfo.failedCount)
                )
            if runInfo.errorCount:
                msgs.append(
                    "<font color=%s>%s errors</font>"
                    % (uiconstants.LOG_COLOR_ERROR.name(), runInfo.errorCount)
                )
            if runInfo.skipCount:
                msgs.append(
                    "<font color=%s>%s skipped</font>"
                    % (uiconstants.LOG_COLOR_WARNING.name(), runInfo.skipCount)
                )
            if runInfo.expectedFailureCount:
                msgs.append(
                    "<font color=%s>%s expected failures</font>"
                    % (
                        uiconstants.LOG_COLOR_SUCCESS.name(),
                        runInfo.expectedFailureCount,
                    )
                )
            if runInfo.unexpectedSuccessCount:
                msgs.append(
                    "<font color=%s>%s unexpected successes</font>"
                    % (
                        uiconstants.LOG_COLOR_ERROR.name(),
                        runInfo.unexpectedSuccessCount,
                    )
                )
        self.setText(", ".join(msgs))

    def reportTestCount(self, testCount):
        self.setText("{} tests.".format(testCount))

    def startCollectingTests(self):
        self.setText("Loading tests...")
        self.repaint()



import logging
import sys

logger = logging.getLogger(__name__)


class _QtModuleImporter(object):
    QtCore = None
    QtGui = None
    QtWidgets = None
    QtSvg = None

    Variant = None
    Signal = None

    IsQt5 = False

    @classmethod
    def _importFromPySide2(cls):
        try:
            from PySide2 import QtCore, QtGui, QtWidgets, QtSvg

            logger.debug("Using PySide2.")
            cls.QtCore = QtCore
            cls.QtGui = QtGui
            cls.QtSvg = QtSvg
            cls.QtWidgets = QtWidgets
            cls.Signal = QtCore.Signal
            cls.IsQt5 = True
        except Exception:
            logger.debug("Unable to import PySide2.")

    @classmethod
    def _importFromPySide1(cls):
        try:
            from PySide import QtCore, QtGui, QtSvg

            logger.debug("Using PySide.")
            cls.QtCore = QtCore
            cls.QtGui = QtGui
            cls.QtWidgets = QtGui
            cls.QtSvg = QtSvg
            cls.Signal = QtCore.Signal
            cls.IsQt5 = False
        except Exception:
            logger.debug("Unable to import PySide.")

    @classmethod
    def _importFromPyQt5(cls):
        try:
            from PyQt5 import QtCore, QtGui, QtWidgets, QtSvg

            logger.debug("Using PyQt5.")
            cls.QtCore = QtCore
            cls.QtGui = QtGui
            cls.QtWidgets = QtWidgets
            cls.QtSvg = QtSvg
            cls.Signal = QtCore.pyqtSignal
            cls.Variant = QtCore.QVariant
            cls.IsQt5 = True
        except Exception:
            logger.debug("Unable to import PyQt5.")

    @classmethod
    def _importFromPyQt4(cls):
        try:
            from PyQt4 import QtCore, QtGui, QtSvg

            logger.debug("Using PyQt4.")
            cls._isPyQt = True
            cls.QtCore = QtCore
            cls.QtGui = QtGui
            cls.QtWidgets = QtGui
            cls.QtSvg = QtSvg
            cls.Signal = QtCore.pyqtSignal
            cls.Variant = QtCore.QVariant
            cls.IsQt5 = False
        except Exception:
            logger.debug("Unable to import PyQt4.")

    @classmethod
    def importModules(cls):
        importers = (
            cls._importFromPySide2,
            cls._importFromPySide1,
            cls._importFromPyQt5,
            cls._importFromPyQt4,
        )
        for importer in importers:
            importer()
            if cls.QtCore:
                return

        logger.debug("No Qt modules found.")


_QtModuleImporter.importModules()
QtCore = _QtModuleImporter.QtCore
QtGui = _QtModuleImporter.QtGui
QtWidgets = _QtModuleImporter.QtWidgets
QtSvg = _QtModuleImporter.QtSvg
Signal = _QtModuleImporter.Signal


def findTopLevelWidgetByName(name):
    for wgt in QtWidgets.QApplication.topLevelWidgets():
        if wgt.objectName() == name:
            return wgt
    return None


def variantToPyValue(value):
    if _QtModuleImporter.Variant and isinstance(value, _QtModuleImporter.Variant):
        return value.toPyObject()
    return value


def iconFromPath(filePath):
    if QtGui != QtWidgets:
        return QtGui.QIcon(filePath)

    renderer = QtSvg.QSvgRenderer(filePath)
    pixmap = QtGui.QPixmap(renderer.defaultSize())
    pixmap.fill(QtCore.Qt.transparent)
    painter = QtGui.QPainter(pixmap)
    renderer.render(painter)
    painter.end()
    return QtGui.QIcon(pixmap)


def isQt5():
    return _QtModuleImporter.IsQt5


def checkQtAvailability():
    if bool(QtCore):
        return True

    if sys.version_info.major >= 3:
        qtLibs = ["PySide2", "PyQt5"]
    else:
        qtLibs = ["PySide"]
    cmdStr = " or ".join(["`pip install {}`".format(lib) for lib in qtLibs])
    print(
        "QT is required by IUTest UI, please install it by {}".format(cmdStr)
    )
    return False


def setDarkStyle():
    presetStyles = ()
    if hasattr(QtGui, "QStyleFactory"):
        presetStyles = QtGui.QStyleFactory.keys()
    else:
        presetStyles = QtWidgets.QStyleFactory.keys()

    if "Fusion" in presetStyles:
        QtWidgets.QApplication.setStyle("Fusion")
    elif "plastique" in presetStyles:
        QtWidgets.QApplication.setStyle("plastique")

    palette = QtGui.QPalette()
    palette.setColor(palette.Window, QtGui.QColor(68, 68, 68))
    palette.setColor(palette.WindowText, QtGui.QColor(200, 200, 200))
    palette.setColor(palette.Base, QtGui.QColor(46, 46, 46))
    palette.setColor(palette.AlternateBase, QtGui.QColor(43, 43, 43))
    palette.setColor(palette.ToolTipBase, QtGui.QColor(46, 46, 46))
    palette.setColor(palette.ToolTipText, QtCore.Qt.white)
    palette.setColor(palette.Text, QtGui.QColor(200, 200, 200))
    palette.setColor(palette.Disabled, palette.Text, QtGui.QColor(120, 120, 120))
    palette.setColor(palette.Button, QtGui.QColor(53, 53, 53))
    palette.setColor(palette.ButtonText, QtGui.QColor(238, 238, 238))
    palette.setColor(palette.Disabled, palette.ButtonText, QtGui.QColor(120, 120, 120))
    palette.setColor(palette.BrightText, QtGui.QColor(238, 238, 238))
    palette.setColor(palette.Link, QtGui.QColor(42, 130, 218))
    palette.setColor(palette.Highlight, QtGui.QColor(82, 133, 166))
    palette.setColor(palette.HighlightedText, QtGui.QColor(238, 238, 238))
    QtWidgets.QApplication.setPalette(palette)


class ApplicationContext(object):
    """Enable widget works both in standalone mode or DCC embedded mode.
    """

    def __init__(self, darkStyle=True, exit_=False):
        self.isStandalone = False
        self._exit = exit_
        self._darkStyle = darkStyle
        self._hadError = False

    def setHasError(self, hasError):
        self._hadError = hasError

    def __enter__(self):
        self._application = QtWidgets.QApplication.instance()
        if not self._application:
            self.isStandalone = True
            self._application = QtWidgets.QApplication(sys.argv)
            if self._darkStyle:
                setDarkStyle()
        else:
            self.isStandalone = (
                hasattr(self._application, "_eventLoopExited")
                and self._application._eventLoopExited
            )
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._hadError:
            return

        if self.isStandalone:
            if self._exit:
                sys.exit(self._application.exec_())
            else:
                self._application.exec_()
                self._application._eventLoopExited = True




import logging

from iutest import dcc
from iutest import _version
from iutest.core import importutils
from iutest.qt import QtCore, QtWidgets, iconFromPath, variantToPyValue
from iutest.core import iconutils
from iutest.core import appsettings
from iutest.core import constants
from iutest.core import testmanager
from iutest.core import uistream
from iutest.ui import logbrowser
from iutest.ui import rootpathedit
from iutest.ui import unittesttree
from iutest.ui import statuslabel
from iutest.ui import uiutils
from iutest.ui import configwindow

logger = logging.getLogger(__name__)


class IUTestWindow(QtWidgets.QWidget):
    _iutestIcon = None
    _reimportIcon = None
    _reloadUiIcon = None
    _clearIcon = None
    _moreIcon = None
    _filterIcon = None
    _autoFilterIcon = None
    _resetIcon = None
    _clearLogOnRunIcon = None
    _clearLogIcon = None
    _stopAtErrorIcon = None
    _configIcon = None
    _panelStateIconSet = None

    def __init__(self, startDirOrModule=None, topDir=None, parent=None):
        parent = parent or dcc.findParentWindow()
        QtWidgets.QWidget.__init__(self, parent)

        self._initIcons()

        self._testManager = testmanager.TestManager(self, startDirOrModule, topDir)

        self._mainLay = uiutils.makeMainLayout(self)
        self.setContentsMargins(0, 0, 0, 0)

        # Top layout ------------------------------
        self._dirLayout = uiutils.makeMinorHorizontalLayout()
        self._testRunnerActions = []
        self._makeDirWidgets()
        self._mainLay.addLayout(self._dirLayout)

        self._splitter = QtWidgets.QSplitter(self)
        self._mainLay.addWidget(self._splitter, 1)

        # left layout -----------------------------------
        self._leftWidget, leftLay = self._createSplitterContent()

        _topLayout = uiutils.makeMinorHorizontalLayout()
        self._makeTreeTopWidgets(_topLayout)
        leftLay.addLayout(_topLayout)

        self._view = unittesttree.UnitTestTreeView(self)
        self._view.runAllTest.connect(self._runAllTests)
        self._view.runTests.connect(self._runTests)
        self._view.runSetupOnly.connect(self._runTestSetupOnly)
        self._view.runWithoutTearDown.connect(self._runTestWithoutTearDown)
        self._view.itemSelectionChanged.connect(self._viewSelectionChanged)

        leftLay.addWidget(self._view)

        # right layout -----------------------------------
        self._rightWidget, rightLay = self._createSplitterContent()
        _logTopLayout = uiutils.makeMinorHorizontalLayout()
        self._logWgt = logbrowser.LogBrowser(self)
        self._makeLogTopLayout(_logTopLayout)
        rightLay.addLayout(_logTopLayout, 0)
        rightLay.addWidget(self._logWgt, 1)

        self._splitter.setSizes([230, 500])
        self._splitter.setStretchFactor(0, 0)
        self._splitter.setStretchFactor(1, 1)

        # bottom -----------------------------------
        self._statusLbl = statuslabel.StatusLabel(self)
        self._mainLay.addWidget(self._statusLbl)

        _btmLayout = uiutils.makeMinorHorizontalLayout()
        self._makeRunButtons(_btmLayout)
        self._mainLay.addLayout(_btmLayout)

        self.resize(QtCore.QSize(900, 500))
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose, True)  # for reimport convenience.
        self.setTabOrder(self._rootDirLE, self._searchLE)
        self.setTabOrder(self._searchLE, self._view)
        self.setTabOrder(self._view, self._logWgt)
        self.setWindowIcon(self._iutestIcon)
        self.setWindowFlags(QtCore.Qt.Window)

        self._updateWindowTitle()
        self._loadLastDirsFromSettings()
        self._restorePanelVisState()

        self.resetUiTestManager()

        QtCore.QTimer.singleShot(0, self._initialLoad)

    def getLogBrowserWidget(self):
        return self._logWgt

    def resetUiTestManager(self):
        self._view.setTestManager(self._testManager)
        self._statusLbl.setTestManager(self._testManager)

    def _initialLoad(self):
        self.reload(keepUiStates=False)
        self._updateReimportRerunButtonEnabled()

    def _createSplitterContent(self):
        wgt = QtWidgets.QWidget(self)
        splitterLay = QtWidgets.QVBoxLayout(wgt)
        splitterLay.setSpacing(3)
        splitterLay.setContentsMargins(0, 0, 0, 0)
        self._splitter.addWidget(wgt)
        return wgt, splitterLay

    def _makeLogTopLayout(self, layout):
        _console = QtWidgets.QLabel("Log Browser")

        self._clearLogBtn = uiutils.makeIconButton(self._clearLogIcon, self)
        self._clearLogBtn.setToolTip("Clear the log browser logging.")
        self._clearLogBtn.clicked.connect(self._logWgt.clear)

        layout.addWidget(_console, 0)
        layout.addStretch(1)
        layout.addWidget(self._clearLogBtn, 0)

    @classmethod
    def _initSingleIcon(cls, iconVarName, iconFileName):
        iconPath = iconutils.iconPath(iconFileName)
        setattr(cls, iconVarName, iconFromPath(iconPath))

    @classmethod
    def _initIcons(cls):
        if cls._panelStateIconSet:
            return

        panelStatePaths = iconutils.iconPathSet(
            "panelMode.svg", constants.PANEL_VIS_STATE_ICON_SUFFIXES
        )
        cls._panelStateIconSet = [iconFromPath(p) for p in panelStatePaths]

        iconutils.initSingleClassIcon(cls, "_iutestIcon", "iutest.svg")
        iconutils.initSingleClassIcon(cls,"_reimportIcon", "reimport.svg")
        iconutils.initSingleClassIcon(cls,"_reloadUiIcon", "reloadUI.svg")
        iconutils.initSingleClassIcon(cls,"_configIcon", "config.svg")
        iconutils.initSingleClassIcon(cls,"_clearIcon", "clear.svg")
        iconutils.initSingleClassIcon(cls,"_moreIcon", "more.svg")
        iconutils.initSingleClassIcon(cls,"_filterIcon", "stateFilter.svg")
        iconutils.initSingleClassIcon(cls,"_autoFilterIcon", "autoFilter.svg")
        iconutils.initSingleClassIcon(cls,"_resetIcon", "reset.svg")
        iconutils.initSingleClassIcon(cls,"_clearLogIcon", "clearLog.svg")
        iconutils.initSingleClassIcon(cls,"_clearLogOnRunIcon", "clearLogOnRun.svg")
        iconutils.initSingleClassIcon(cls,"_stopAtErrorIcon", "stopAtError.svg")

    def _addToggleConfigAction(self, lbl, icon, tooltip, configKey, slot):
        # Cannot use self._configMenu.addAction() directly here since over-zealous garbage collection in some DCC apps:
        act = QtWidgets.QAction(lbl, self)
        act.setIcon(icon)
        act.setCheckable(True)
        act.setToolTip(tooltip)
        act.toggled.connect(slot)
        value = appsettings.get().simpleConfigBoolValue(configKey)
        act.setChecked(value)
        self._configMenu.addAction(act)
        return (act, value)

    def _regenerateBrowseMenu(self):
        self._browseMenu.clear()

        act = self._browseMenu.addAction("Pick Test Root Dir ...")
        act.triggered.connect(self._onBrowseTestsRootDir)
        act = self._browseMenu.addAction("Pick Top Root Dir ...")
        act.triggered.connect(self._onBrowseTopDir)
        self._browseMenu.addSeparator()
        act = self._browseMenu.addAction("Save Current Path Settings ...")
        act.triggered.connect(self._onSaveCurrentDirSettings)
        act = self._browseMenu.addAction("Delete Current Settings ...")
        act.triggered.connect(self._onDeleteCurrentDirSettings)
        self._browseMenu.addSeparator()

        # Read settings:
        config = appsettings.get().testDirSettings()
        for key, pair in config.items():
            act = self._browseMenu.addAction(key)
            act.triggered.connect(self._loadSavedDirPair)
            act.setToolTip("\n".join(pair))
        self._browseMenu.addSeparator()

    def _updateLabelOfTestRunnerAct(self, act, currentMode):
        runnerMode = variantToPyValue(act.data())
        runner = self._testManager.getRunnerByMode(runnerMode)
        suffix = " [actived]" if runnerMode == currentMode else ""
        lbl = runner.name() + suffix
        act.setText(lbl)

    def _updateConfigButton(self, runner):
        self._runnerConfigBtn.setIcon(runner.icon())
        self._runnerConfigBtn.setToolTip("Test Runner: {}, click for more settings.".format(runner.name()))

    def _makeRunnerConfigButton(self):
        currentRunner = self._setInitialTestMode()
        currentRunnerMode = currentRunner.mode()
        self._runnerConfigBtn, self._configMenu = self._makeMenuToolButton()
        self._updateConfigButton(currentRunner)
        
        for runner in self._testManager.iterAllRunners():
            icon = runner.icon()
            toolTip = "Run tests using {}.".format(runner.name())
            runnerMode = runner.mode()
            act = QtWidgets.QAction(runner.name(), self)
            act.setIcon(icon)
            act.setEnabled(runner.isValid())
            act.setToolTip(toolTip)
            act.setData(runnerMode)
            act.triggered.connect(self._onTestRunnerSwitched)
            self._updateLabelOfTestRunnerAct(act, currentRunnerMode)
            self._testRunnerActions.append(act)
            self._configMenu.addAction(act)

        self._configMenu.addSeparator()

        # Stop on error act:
        self._stopOnErrorAct, stopOnError = self._addToggleConfigAction(
            "Stop On Error",
            self._stopAtErrorIcon,
            "Stop the tests running once it encounters a test error/failure.",
            configKey=constants.CONFIG_KEY_STOP_ON_ERROR,
            slot=self._onStopOnErrorActionToggled
        )
        self._testManager.setStopOnError(stopOnError)

        # auto filer on run act:
        self._autoFilterAct, stopOnError = self._addToggleConfigAction(
            "Only Show Tests that Run",
            self._autoFilterIcon,
            "Apply a ':ran' filter automatically after every test run so that only the run tests shown in the view.",
            configKey=constants.CONFIG_KEY_AUTO_FILTERING_STATE,
            slot=self._onAutoFilterActionToggled
        )

        # auto clear log on run act:
        self._clearLogOnRunAct, stopOnError = self._addToggleConfigAction(
            "Clear LogBrowser On Run",
            self._clearLogOnRunIcon,
            "Clear the log browser automatically before every test run.",
            configKey=constants.CONFIG_KEY_AUTO_CLEAR_LOG_STATE,
            slot=self._onAutoClearLogActionToggled
        )

        self._configMenu.addSeparator()
        act = self._configMenu.addAction("Preference..")
        act.setIcon(self._configIcon)
        act.triggered.connect(self._showConfigWindow)

    def _showConfigWindow(self):
        configwindow.ConfigWindow.show(self)

    def _makeMenuToolButton(self, icon=None, toolTip=""):
        btn = QtWidgets.QToolButton(self)
        if icon:
            btn.setIcon(icon)
        btn.setToolTip(toolTip)
        btn.setContentsMargins(0,0,0,0)
        btn.setIconSize(QtCore.QSize(20,20))
        btn.setPopupMode(btn.InstantPopup)
        menu = QtWidgets.QMenu(btn)
        btn.setMenu(menu)
        return (btn, menu)

    def _setInitialTestMode(self):
        initRunnerMode = appsettings.get().simpleConfigIntValue(
            constants.CONFIG_KEY_LAST_RUNNER_MODE, -1
        )
        if initRunnerMode < 0:
            initRunnerMode = self._testManager.getFeasibleRunnerMode()
            
        self._testManager.setRunnerMode(initRunnerMode)
        return self._testManager.getRunner()

    def _makeDirWidgets(self):
        lbl = QtWidgets.QLabel("Test Root")
        lbl.setToolTip(
            "Input a python module path or an absolute dir path in the lineEdit for the tests."
        )
        self._rootDirLE = rootpathedit.RootPathEdit(self)
        self._rootDirLE.rootPathChanged.connect(self._switchToTestRootPath)

        self._browseBtn, self._browseMenu = self._makeMenuToolButton(
            self._moreIcon, "Click to access more features."
        )
        self._regenerateBrowseMenu()

        self._panelVisBtn = uiutils.makeIconButton(self._panelStateIconSet[-1], self)
        self._panelVisBtn.setToolTip(
            "Switch the test tree view and the log browser visibility."
        )
        self._panelVisBtn.clicked.connect(self._onPanelVisButtonClicked)
        self._makeRunnerConfigButton()

        self._dirLayout.addWidget(lbl)
        self._dirLayout.addWidget(self._rootDirLE)
        self._dirLayout.addWidget(self._browseBtn)
        self._dirLayout.addWidget(self._panelVisBtn)
        self._dirLayout.addWidget(self._runnerConfigBtn)

        self._updateDirUI()

    def _makeTreeTopWidgets(self, layout):
        reimportBtn = uiutils.makeIconButton(self._reimportIcon, self)
        reimportBtn.setVisible(importutils.isReimportFeatureAvailable(silentCheck=True))
        reimportBtn.setToolTip("Reimport all changed python module.")
        reimportBtn.clicked.connect(self._reimportAllChangedModules)
        layout.addWidget(reimportBtn)

        reloadUIBtn = uiutils.makeIconButton(self._reloadUiIcon, self)
        reloadUIBtn.setToolTip("Recollect tests and reload the test tree view below.")
        reloadUIBtn.clicked.connect(self._onReloadUiButtonClicked)
        layout.addWidget(reloadUIBtn)

        self._searchLE = QtWidgets.QLineEdit()
        self._searchLE.setToolTip(
            "Input keywords to filter the tests, separated by space.\n"
            "For normal keyword, the match operation is 'AND'\n"
            "For state keyword starting with ':', the match operation is 'OR'."
        )
        self._searchLE.setPlaceholderText("Input to filter")
        self._searchLE.textChanged.connect(self._applyFilterText)
        layout.addWidget(self._searchLE)

        self._clearFilterBtn = uiutils.makeIconButton(self._clearIcon, self)
        self._clearFilterBtn.setToolTip("Clear all filters.")
        self._clearFilterBtn.setEnabled(False)
        self._clearFilterBtn.clicked.connect(self._clearFilter)
        layout.addWidget(self._clearFilterBtn)

        self._filterBtn, self._filterMenu = self._makeMenuToolButton(
            self._filterIcon, "Click to apply more predefined filters."
        )
        for lbl in constants.KEYWORD_TEST_STATES:
            act = self._filterMenu.addAction(lbl)
            act.triggered.connect(self._addStateFilter)
        layout.addWidget(self._filterBtn)

    def _onAutoFilterActionToggled(self, state):
        appsettings.get().saveSimpleConfig(
            constants.CONFIG_KEY_AUTO_FILTERING_STATE, state
        )

    def _onAutoClearLogActionToggled(self, state):
        appsettings.get().saveSimpleConfig(
            constants.CONFIG_KEY_AUTO_CLEAR_LOG_STATE, state
        )

    def _onTestRunnerSwitched(self):
        act = self.sender()
        if not act:
            return

        runnerMode = variantToPyValue(act.data())
        self._testManager.setRunnerMode(runnerMode)
        runner = self._testManager.getRunner()
        self._updateConfigButton(runner)
        logger.info("Switch to %s for test running.", runner.name())
        
        self._onReloadUiButtonClicked()
        appsettings.get().saveSimpleConfig(
            constants.CONFIG_KEY_LAST_RUNNER_MODE, runnerMode
        )
        for act in self._testRunnerActions:
            self._updateLabelOfTestRunnerAct(act, runnerMode)

    def _onStopOnErrorActionToggled(self, stop):
        self._testManager.setStopOnError(stop)
        appsettings.get().saveSimpleConfig(constants.CONFIG_KEY_STOP_ON_ERROR, stop)

    def _makeRunButtons(self, _btmLayout):
        self._resetAllBtn = uiutils.makeIconButton(self._resetIcon, self)
        self._resetAllBtn.setToolTip(
            "Reset the test item states, icons and stats and collapse the tree view items in a certain level."
        )
        self._resetAllBtn.clicked.connect(self._view.resetAllItemsToNormal)
        _btmLayout.addWidget(self._resetAllBtn, 1)

        self._runSelectedBtn = QtWidgets.QPushButton("Run &Selected", self)
        self._runSelectedBtn.setToolTip("Run the selected tests in the view.")
        self._runSelectedBtn.clicked.connect(self._runViewSelectedTests)
        self._runSelectedBtn.setIcon(self._view._runSelectedIcon)
        _btmLayout.addWidget(self._runSelectedBtn, 1)

        self._runAllBtn = QtWidgets.QPushButton("Run &All", self)
        self._runAllBtn.setToolTip(
            "Run all the tests, including those filtered from the view."
        )
        self._runAllBtn.setIcon(self._view._runAllIcon)
        self._runAllBtn.clicked.connect(self._runAllTests)
        _btmLayout.addWidget(self._runAllBtn, 1)

        self._reimportAndRerunBtn = QtWidgets.QPushButton("&Reload && Rerun", self)
        self._reimportAndRerunBtn.setToolTip(
            "Reimport all changed python modules and rerun the last tests."
        )
        self._reimportAndRerunBtn.clicked.connect(self._reimportPyAndRerun)
        self._reimportAndRerunBtn.setIcon(self._view._reimportAndRunIcon)
        _btmLayout.addWidget(self._reimportAndRerunBtn, 1)

        _runMoreBtn = self._makeRunMoreButton()
        _btmLayout.addWidget(_runMoreBtn, 0)

    def _makeRunMoreButton(self):
        _runMoreBtn = QtWidgets.QToolButton(self)
        _runMoreBtn.setIcon(self._view._runPartialIcon)
        _runMoreBtn.setAutoRaise(True)
        _runMoreBtn.setToolButtonStyle(QtCore.Qt.ToolButtonIconOnly)
        _runMoreBtn.setPopupMode(_runMoreBtn.InstantPopup)

        self._runSetupAct = QtWidgets.QAction("Run setUp( ) Only", self)
        self._runSetupAct.setIcon(self._view._runPartialIcon)
        self._runSetupAct.triggered.connect(self._runTestSetupOnly)

        self._runNoTearDown = QtWidgets.QAction("Run without tearDown( )", self)
        self._runNoTearDown.setIcon(self._view._runPartialIcon)
        self._runNoTearDown.triggered.connect(self._runTestWithoutTearDown)

        menu = QtWidgets.QMenu(_runMoreBtn)
        menu.addAction(self._runSetupAct)
        menu.addAction(self._runNoTearDown)
        _runMoreBtn.setMenu(menu)
        return _runMoreBtn

    def _reimportAllChangedModules(self):
        importutils.reimportAllChangedPythonModules()

    def _onReloadUiButtonClicked(self):
        self.reload(keepUiStates=True)
        self._applyCurrentFilter(removeStateFilters=True, keepUiStates=True)

    def _addStateFilter(self):
        stateKeyword = str(self.sender().text())
        search = str(self._searchLE.text()).strip()
        if not search:
            self._searchLE.setText(stateKeyword)
            return
        parts = search.split(" ")
        if stateKeyword in parts:
            return
        parts.insert(0, stateKeyword)
        self._searchLE.setText(" ".join(parts))

    def _saveLastTestDir(self, startDir, topDir):
        settings = appsettings.get()
        settings.saveSimpleConfig(
            constants.CONFIG_KEY_LAST_TEST_ROOT_DIR, startDir, sync=False
        )
        settings.saveSimpleConfig(constants.CONFIG_KEY_LAST_TOP_DIR, topDir, sync=True)
        logger.info("Save the last test root: %s", startDir)

    def _loadSavedDirPair(self):
        act = self.sender()
        _topDir, _startDirOrModule = act.toolTip().split("\n")
        self._testManager.setTopDir(_topDir)
        self._testManager.setStartDirOrModule(_startDirOrModule)

        configName = str(act.text())
        self._saveLastTestDir(_startDirOrModule, _topDir)
        self._updateWindowTitle(configName)
        self._updateDirUI()
        self._searchLE.clear()
        self.reload()

    def _switchToTestRootPath(self, startDir, topDir):
        self._testManager.setDirs(startDir, topDir)
        self._updateWindowTitle(startDir)
        self._updateDirUI()
        self._searchLE.clear()
        self.reload(keepUiStates=False)
        if not self._testManager.hasLastListerError():
            self._saveLastTestDir(startDir, topDir)

    def _setPanelVisState(self, state, saveSettings=True):
        state = min(constants.PANEL_VIS_STATE_BOTH_ON, max(0, int(state)))
        self._panelVisBtn.setIcon(self._panelStateIconSet[state])
        self._leftWidget.setVisible(state != constants.PANEL_VIS_STATE_RIGHT_ON)
        self._rightWidget.setVisible(state != constants.PANEL_VIS_STATE_LEFT_ON)
        if saveSettings:
            appsettings.get().saveSimpleConfig(
                constants.CONFIG_KEY_PANEL_VIS_STATE, state
            )

    def _restorePanelVisState(self):
        state = appsettings.get().simpleConfigIntValue(
            constants.CONFIG_KEY_PANEL_VIS_STATE, constants.PANEL_VIS_STATE_BOTH_ON
        )
        self._setPanelVisState(state, saveSettings=False)

    def _onPanelVisButtonClicked(self):
        leftVis = self._leftWidget.isVisible()
        rightVis = self._rightWidget.isVisible()
        state = constants.PANEL_VIS_STATE_BOTH_ON
        if not leftVis:
            state = constants.PANEL_VIS_STATE_RIGHT_ON
        elif not rightVis:
            state = constants.PANEL_VIS_STATE_LEFT_ON
        state = (state + 1) % 3
        self._setPanelVisState(state)

    def _updateDirUI(self):
        self._rootDirLE.setText(self._testManager.startDirOrModule())
        self._rootDirLE.setToolTip(self._testManager.topDir())

    def _onBrowseTestsRootDir(self):
        dirPath = QtWidgets.QFileDialog.getExistingDirectory(
            self, "Pick the Test Root Directory"
        )
        if dirPath:
            self._testManager.setStartDirOrModule(dirPath)
            self._updateDirUI()
            self._searchLE.clear()
            self.reload()

    def _onBrowseTopDir(self):
        dirPath = QtWidgets.QFileDialog.getExistingDirectory(
            self, "Pick the Top Python Directory"
        )
        if dirPath:
            self._testManager.setTopDir(dirPath)

        self._searchLE.clear()
        self._updateDirUI()
        self.reload()

    def _updateWindowTitle(self, configName=None):
        if configName:
            self.setWindowTitle("{} {} - {}".format(constants.APP_NAME, _version.__version__, configName))
        else:
            self.setWindowTitle("{} {}".format(constants.APP_NAME, _version.__version__))

    def _configNameFromStartAndTopDir(self, startDir, topDir):
        config = appsettings.get().testDirSettings()
        for name in config:
            _topDir, _startDirOrModule = config[name]
            if _topDir == topDir and _startDirOrModule == startDir:
                return name
        return startDir

    def _loadLastDirsFromSettings(self):
        settings = appsettings.get()
        startDirOrModule = settings.simpleConfigStrValue(
            constants.CONFIG_KEY_LAST_TEST_ROOT_DIR
        )
        if not startDirOrModule:
            return

        topDir = settings.simpleConfigStrValue(constants.CONFIG_KEY_LAST_TOP_DIR)

        self._testManager.setDirs(startDirOrModule, topDir)
        self._updateDirUI()
        self._updateWindowTitle(
            self._configNameFromStartAndTopDir(startDirOrModule, topDir)
        )

    def _onSaveCurrentDirSettings(self):
        startDir = self._testManager.startDirOrModule()
        name = QtWidgets.QInputDialog.getText(
            self, "Save Dir Settings", "Input a name for the test set:", text=startDir
        )

        if not name[-1] or not name[0]:
            return

        topDir = self._testManager.topDir()
        with appsettings.SettingsGroupContext(constants.CONFIG_KEY_SAVED_TEST_DIR):
            with appsettings.SettingsGroupContext(name[0]) as settings:
                settings.saveSimpleConfig(
                    constants.CONFIG_KEY_TEST_TOP_DER, topDir, sync=False
                )
                settings.saveSimpleConfig(
                    constants.CONFIG_KEY_TEST_START_DER, startDir, sync=False
                )

        self._saveLastTestDir(startDir, topDir)
        self._deferredRegenerateMenu()
        self._updateWindowTitle(name[0])

    def _deferredRegenerateMenu(self):
        QtCore.QTimer.singleShot(0, self._regenerateBrowseMenu)

    def _onDeleteCurrentDirSettings(self):
        config = appsettings.get().testDirSettings()
        removeNames = []
        startDir = self._testManager.startDirOrModule()
        for name, pair in config.items():
            _, root = pair
            if root == startDir:
                removeNames.append(name)
        if not removeNames:
            logger.warning("Current dir settings was not saved before.")
            return

        msg = "Are you sure to delete these dir settings: %s" % removeNames
        answer = QtWidgets.QMessageBox.question(self, "Remove Current Dir Setting", msg)
        if answer != QtWidgets.QMessageBox.Yes:
            return

        with appsettings.SettingsGroupContext(
            constants.CONFIG_KEY_SAVED_TEST_DIR
        ) as settings:
            for n in removeNames:
                settings.removeConfig(n)
                logger.info("Remove the dir setting: %s", n)

        self._deferredRegenerateMenu()

    def _runAllTests(self):
        self._view.resetAllItemsToNormal()
        self._searchLE.clear()
        self._testManager.runAllTests()
        self._updateReimportRerunButtonEnabled()

    def _runTests(self, testIds):
        if not testIds:
            return

        self._beforeRunningTests(testIds)
        self._testManager.runTests(*testIds)
        self._updateReimportRerunButtonEnabled()

    def _viewSelectionChanged(self):
        hasSel = self._view.hasSelectedTests()
        self._runSelectedBtn.setEnabled(hasSel)
        hasSelForPartialRun = self._view.hasSelectedTests(hasSelectedTestOrCase=True)
        self._runSetupAct.setEnabled(hasSelForPartialRun)
        self._runNoTearDown.setEnabled(hasSelForPartialRun)

    def _runViewSelectedTests(self):
        self._runTests(self._view.selectedTestIds(decomposePackageIfNecessary=True))

    def _updateReimportRerunButtonEnabled(self):
        self._reimportAndRerunBtn.setEnabled(bool(self._testManager.lastRunTestIds()))

    def _reimportPyAndRerun(self):
        importutils.reimportAllChangedPythonModules()
        lastRunIds = self._testManager.lastRunTestIds()
        if not lastRunIds:
            logger.warning("Didn't find the last run tests.")
            return

        self._runTests(lastRunIds)

    def _runTestSetupOnly(self):
        self._runPartialTest(constants.RUN_TEST_SETUP_ONLY)

    def _runTestWithoutTearDown(self):
        self._runPartialTest(constants.RUN_TEST_NO_TEAR_DOWN)

    def _runPartialTest(self, partialMode):
        testId = self._view.firstSelectedTestCaseId()
        if not testId:
            logger.error("You need to select test suite or test case.")
            return
        self._testManager.runSingleTestPartially(testId, partialMode)
        self._updateReimportRerunButtonEnabled()

    def _clearFilter(self):
        self._searchLE.clear()

    def reload(self, keepUiStates=True):
        self._beforeTestCollection()
        with uistream.LogCaptureContext():  # Report the possible error report to UI as well:
            self._view.reload(keepUiStates=keepUiStates)

        self._updateRunButtonsEnabled()

        if not self._testManager.startDirOrModule():
            return

        self._statusLbl.reportTestCount(self._view.testCount())

    def _applyCurrentFilter(self, removeStateFilters=True, keepUiStates=True):
        searchText = str(self._searchLE.text())
        if removeStateFilters:
            keywords = searchText.split(" ")
            filterFunc = lambda x: not x.startswith(":")
            keywords = filter(filterFunc, keywords)
            searchText = " ".join(keywords)
            self._searchLE.setText(searchText)

        self._applyFilterTextWithState(searchText, keepUiStates=keepUiStates)

    def _updateRunButtonsEnabled(self):
        enabled = self._view.hasTests()
        self._resetAllBtn.setEnabled(enabled)
        self._runAllBtn.setEnabled(enabled)
        self._viewSelectionChanged()
        self._updateReimportRerunButtonEnabled()

    def _hookUiToStream(self):
        uistream.UiStream.setUi(self)

    def _beforeTestCollection(self):
        self._hookUiToStream()
        self._statusLbl.startCollectingTests()

    def _beforeRunningTests(self, tests):
        self._hookUiToStream()
        self._view.resetTestItemsById(tests)

    def _applyFilterText(self, txt):
        self._applyFilterTextWithState(str(txt), keepUiStates=False)

    def _applyFilterTextWithState(self, txt, keepUiStates=True):
        lowerTxt = txt.strip().lower()
        keywords = lowerTxt.split(" ")
        self._clearFilterBtn.setEnabled(bool(keywords))
        self._view.setFilterKeywords(keywords, ensureFirstMatchVisible=not keepUiStates)
    
    def closeEvent(self, event):
        uistream.UiStream.unsetUi(self)
        QtWidgets.QWidget.closeEvent(self, event)

    # These below are to be called by hooks -------------------------------------------------
    def onSingleTestStart(self, testId, startTime):
        self._hookUiToStream()
        self._view.onSingleTestStart(testId, startTime)

    def onSingleTestStop(self, testId, endTime):
        self._view.onSingleTestStop(testId, endTime)
        self._statusLbl.updateReport()

    def showResultOnItemByTestId(self, testId, state):
        self._view.showResultOnItemByTestId(testId, state)

    def onTestRunningSessionStart(self):
        self._statusLbl.setText("Running tests...")
        if self._clearLogOnRunAct.isChecked():
            self._logWgt.clear()
        self._logWgt.logSeparator()

    def onAllTestsFinished(self):
        self._view.onAllTestsFinished()
        if self._autoFilterAct.isChecked():
            self._searchLE.setText(constants.KEYWORD_TEST_STATE_RUN)
        self._statusLbl.updateReport()

    def repaintUi(self):
        eventFlags = (
            QtCore.QEventLoop.ExcludeSocketNotifiers
            | QtCore.QEventLoop.ExcludeUserInputEvents
        )
        QtWidgets.QApplication.processEvents(eventFlags)


if __name__ == "__main__":
    manager = IUTestWindow()
    manager.show()



from setuptools import setup


def readme():
    src = "./iutest/icons/iutest.svg"
    dst = "http://mgland.com/shared/iutest.svg"
    with open("README.md") as f:
        longDesc = f.read()
    return longDesc.replace(src, dst)


def version():
    with open("iutest/_version.py") as f:
        return f.readlines()[-1].split()[-1].strip("\"'")


setup(
    name="iutest",
    version=version(),
    description="An interactive python test runner supports various test frameworks.",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/mgland/iutest",
    author="Miguel Gao",
    author_email="opensourcemg@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Framework :: Pytest",
        "Topic :: Software Development",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Testing :: Unit",
        "Intended Audience :: Developers",
    ],
    packages=[
        "iutest",
        "iutest.core",
        "iutest.core.runners",
        "iutest.dcc",
        "iutest.docs",
        "iutest.plugins",
        "iutest.plugins.nose2plugins",
        "iutest.plugins.pyunitextentions",
        "iutest.tests",
        "iutest.tests.empty",
        "iutest.tests.iutests",
        "iutest.ui",
    ],
    include_package_data=True,
    install_requires=[],
    entry_points={"console_scripts": ["iutest=iutest.cli:main"]},
)



import logging
from iutest import dependencies
from iutest.core import pathutils
from iutest.core import runinfo
from iutest.core.runners import base
from iutest.core.runners import runnerconstants

logger = logging.getLogger(__name__)


class Nose2TestRunner(base.BaseTestRunner):
    uihooks = None
    testlister = None
    partialtest = None

    @classmethod
    def _importPlugins(cls):
        if cls.uihooks and cls.testlister and cls.partialtest:
            return True

        try:
            from iutest.plugins.nose2plugins import uihooks
            from iutest.plugins.nose2plugins import testlister
            from iutest.plugins.nose2plugins import partialtest

            cls.uihooks = uihooks
            cls.testlister = testlister
            cls.partialtest = partialtest
            return True
        except:
            cls.uihooks = None
            cls.testlister = None
            cls.partialtest = None
            return False

    @classmethod
    def mode(cls):
        return runnerconstants.RUNNER_NOSE2

    @classmethod
    def isValid(cls):
        return dependencies.Nose2Wrapper.get().isValid()

    @classmethod
    def iconFileName(cls):
        return "nose2.svg"

    def runTests(self, *testIds):
        if not self._importPlugins():
            self._issueNotInstalledError()
            return

        plugins = [
            "iutest.plugins.nose2plugins.uihooks",
            "nose2.plugins.loader.eggdiscovery",
            "iutest.plugins.nose2plugins.removeduplicated",
        ]
        excludePlugins = [
            "iutest.plugins.nose2plugins.testlister",
            "iutest.plugins.nose2plugins.partialtest",
        ]
        self._runTest(plugins, excludePlugins, [], *testIds)

    def _runTest(self, plugins, excludePlugins, extraArgs, *testIds):
        if not testIds:
            logger.warning("No test to run.")
            return

        argv = ["nose2", "-v"]
        for plugin in plugins:
            argv.append("--plugin")
            argv.append(plugin)

        for plugin in excludePlugins:
            argv.append("--exclude-plugin")
            argv.append(plugin)

        argv.extend(extraArgs)

        startDirOrModule = self._manager.startDirOrModule()
        topDir = self._manager.topDir()
        # Add start and top dir to avoid potential module name conflict:
        if pathutils.isPath(startDirOrModule):
            argv.extend(["-s", startDirOrModule, "-t", topDir])

        argv.extend(testIds)
        argv.extend(["--fail-fast"] if self._manager.stopOnError() else [])
        self.uihooks.UiHooksPlugin.resetLastData()        
        dependencies.Nose2Wrapper.getModule().discover(
            argv=argv,
            exit=False,
        )

    def runSingleTestPartially(self, testId, partialMode):
        """Run partial steps of test, like running setUp only, or setUp and test but without teardown.
        Args:
            testId (str): id of test, the python module.
            partialMode (int): the test run mode, available values are:
                constants.RUN_TEST_SETUP_ONLY | constants.RUN_TEST_NO_TEAR_DOWN
        """
        if not self._importPlugins():
            self._issueNotInstalledError()
            return

        plugins = [
            "iutest.plugins.nose2plugins.uihooks",
            "nose2.plugins.loader.eggdiscovery",
            "iutest.plugins.nose2plugins.removeduplicated",
            "iutest.plugins.nose2plugins.partialtest",
        ]
        excludePlugins = [
            "iutest.plugins.nose2plugins.testlister",
            "nose2.plugins.result",
        ]
        extraArgs = ["--partial-test"]
        self.partialtest.PartialTestRunner.setRunMode(partialMode)
        self._runTest(plugins, excludePlugins, extraArgs, testId)

    def iterAllTestIds(self):
        if not self._importPlugins():
            self._issueNotInstalledError()
            return

        startDirOrModule = self._manager.startDirOrModule()
        topDir = self._manager.topDir()
        for tid in self.testlister.iterAllTestPathsFromRootDir(
            startDirOrModule, topDir
        ):
            yield tid

    @classmethod
    def hasLastListerError(cls):
        if not cls._importPlugins():
            return False

        return cls.testlister.TestCollector.gotError

    @classmethod
    def lastRunInfo(cls):
        if not cls._importPlugins():
            return runinfo.TestRunInfo()
        return cls.uihooks.UiHooksPlugin.lastRunInfo



from iutest.qt import QtGui

LOG_COLOR_INFORMATION = QtGui.QColor(200, 200, 200)
LOG_COLOR_ERROR = QtGui.QColor(234, 52, 95)
LOG_COLOR_FAILED = QtGui.QColor(234, 52, 95)
LOG_COLOR_WARNING = QtGui.QColor(220, 206, 135)
LOG_COLOR_SUCCESS = QtGui.QColor(138, 211, 11)
